from recon.core.web.utils import columnize
import json
import os

class Tasks(object):

    def __init__(self, recon):
        self.recon = recon
        self.path = os.path.join(self.recon.workspace, 'tasks.db')
        if not os.path.exists(self.path):
            self._create_db()

    def query(self, *args, **kwargs):
        return self.recon._query(self.path, *args, **kwargs)

    def _create_db(self):
        self.query('''\
            CREATE TABLE IF NOT EXISTS tasks (
                id TEXT NOT NULL,
                status TEXT NOT NULL,
                result TEXT,
                PRIMARY KEY (id)
            )'''
        )

    def get_tasks(self):
        tasks = []
        task_ids = [x[0] for x in self.query('SELECT id FROM tasks')]
        for task_id in task_ids:
            tasks.append(self.get_task(task_id))
        return tasks

    def get_ids(self):
        return [t['id'] for t in self.get_tasks()]

    def add_task(self, tid, status, result=None):
        if result:
            result = json.dumps(result)
        return self.query(
            'INSERT INTO tasks (id, status, result) VALUES (?, ?, ?)',
            values=(tid, status, result),
        )

    def get_task(self, tid):
        rows = self.query(
            'SELECT * FROM tasks WHERE id=?',
            values=(tid,),
            include_header=True,
        )
        columns = rows.pop(0)
        task = columnize(columns, rows)[0]
        if task['result']:
            task['result'] = json.loads(task['result'])
        return task

    def update_task(self, tid, **kwargs):
        if kwargs.get('result'):
            kwargs['result'] = json.dumps(kwargs['result'])
        set_items = []
        set_values = []
        for key, value in kwargs.items():
            if kwargs[key]:
                set_items.append(key+'=?')
                set_values.append(value)
        set_str = ', '.join(set_items)
        set_values.append(tid)
        return self.query(
            'UPDATE tasks SET {} WHERE id=?'.format(set_str),
            values=set_values,
        )



from queue import Queue, Empty
import threading
import time

class ThreadingMixin(object):

    def _thread_wrapper(self, *args):
        ''' Wrapper for the worker method defined in the module. Handles calling the actual worker, cleanly exiting upon
        interrupt, and passing exceptions back to the main process.'''
        thread_name = threading.current_thread().name
        self.debug(f"THREAD => {thread_name} started.")
        while not self.stopped.is_set():
            try:
                # use the get_nowait() method for retrieving a queued item to
                # prevent the thread from blocking when the queue is empty
                obj = self.q.get_nowait()
            except Empty:
                continue
            try:
                # launch the public module_thread method
                self.module_thread(obj, *args)
            except:
                # handle exceptions local to the thread
                self.print_exception(f"(thread={thread_name}, object={repr(obj)})")
            finally:
                self.q.task_done()
        self.debug(f"THREAD => {thread_name} exited.")

    # sometimes a keyboardinterrupt causes a race condition between when the self.q.task_done() call above and the
    # self.q.empty() call below, causing all the threads to hang. introducing the time.sleep(.7) call below reduces
    # the likelihood of encountering the race condition.

    def thread(self, *args):
        # disable threading in debug mode
        if self._global_options['verbosity'] >= 2:
            # call the thread method in serial for each input
            for item in args[0]:
                self.module_thread(item, *args[1:])
            return
        # begin threading code
        thread_count = self._global_options['threads']
        self.stopped = threading.Event()
        self.exc_info = None
        self.q = Queue()
        # populate the queue from the user-defined iterable. should be done
        # before the threads start so they have something to process right away
        for item in args[0]:
            self.q.put(item)
        # launch the threads
        threads = []
        for i in range(thread_count):
            t = threading.Thread(target=self._thread_wrapper, args=args[1:])
            threads.append(t)
            t.setDaemon(True)
            t.start()
        # hack to catch keyboard interrupts
        try:
            while not self.q.empty():
                time.sleep(.7)
        except KeyboardInterrupt:
            self.error('Ok. Waiting for threads to exit...')
            # set the event flag to trigger an exit for all threads (interrupt condition)
            self.stopped.set()
            # prevent the module from returning to the interpreter until all threads have exited
            for t in threads:
                t.join()
            raise
        self.q.join()
        # set the event flag to trigger an exit for all threads (normal condition)
        # the threads are no longer needed once all the data has been processed
        self.stopped.set()



from recon.core.web.exports import _jsonify, csvify, xmlify, listify, xlsxify, proxify
from recon.core.web.reports import xlsx, pushpin

EXPORTS = {
    'json': _jsonify,
    'xml': xmlify,
    'csv': csvify,
    'list': listify,
    'xlsx': xlsxify,
    'proxy': proxify,
}

REPORTS = {
    'xlsx': xlsx,
    'pushpin': pushpin,
}



import dns

class ResolverMixin(object):

    def get_resolver(self):
        '''Returns a dnspython default resolver object configured with the framework's global options.'''
        resolver = dns.resolver.Resolver(configure=False)
        resolver.nameservers = [self._global_options['nameserver']]
        resolver.lifetime = 3
        return resolver



import time


class GithubMixin(object):

    def query_github_api(self, endpoint, payload={}, options={}):
        opts = {'max_pages': None}
        opts.update(options)
        headers = {'Authorization': f"token {self.get_key('github_api')}"}
        base_url = 'https://api.github.com'
        url = base_url + endpoint
        results = []
        page = 1
        while True:
            # Github rate limit is 30 requests per minute
            time.sleep(2) # 60s / 30r = 2s/r
            payload['page'] = page
            resp = self.request('GET', url, headers=headers, params=payload)
            # check for errors
            if resp.status_code != 200:
                # skip 404s returned for no results
                if resp.status_code != 404:
                    self.error(f"Message from Github: {resp.json()['message']}")
                break
            # some APIs return lists, and others a single dictionary
            method = 'extend'
            if type(resp.json()) == dict:
                method = 'append'
            getattr(results, method)(resp.json())
            # paginate
            if 'link' in resp.headers and 'rel="next"' in resp.headers['link'] and (opts['max_pages'] is None or page < opts['max_pages']):
                page += 1
                continue
            break
        return results

    def search_github_api(self, query):
        self.verbose(f"Searching Github for: {query}")
        results = self.query_github_api(endpoint='/search/code', payload={'q': query})
        # reduce the nested lists of search results and return
        results = [result['items'] for result in results]
        return [x for sublist in results for x in sublist]



from flask import g, current_app
from sqlite3 import dbapi2 as sqlite3
import os
import re

def columnize(columns, rows):
    return [{columns[i]: row[i] for i in range(0, len(columns))} for row in rows]

def add_worksheet(workbook, name, rows):
    '''Helper function for building xlsx files.'''
    worksheet = workbook.add_worksheet(name)
    # build the data set
    if rows:
        _rows = [list(rows[0].keys())]
        for row in rows:
            _row = []
            for key in _rows[0]:
                _row.append(row[key])
            _rows.append(_row)
        # write the rows of data to the xlsx file
        for r in range(0, len(_rows)):
            for c in range(0, len(_rows[r])):
                worksheet.write(r, c, _rows[r][c])

def is_url(s):
    if type(s) not in (str, bytes):
        return False
    ip_middle_octet = r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))"
    ip_last_octet = r"(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))"
    regex = re.compile(
        r"^"
        # protocol identifier
        r"(?:(?:https?|ftp)://)"
        # user:pass authentication
        r"(?:\S+(?::\S*)?@)?"
        r"(?:"
        r"(?P<private_ip>"
        # IP address exclusion
        # private & local networks
        r"(?:(?:10|127)" + ip_middle_octet + "{2}" + ip_last_octet + ")|"
        r"(?:(?:169\.254|192\.168)" + ip_middle_octet + ip_last_octet + ")|"
        r"(?:172\.(?:1[6-9]|2\d|3[0-1])" + ip_middle_octet + ip_last_octet + "))"
        r"|"
        # IP address dotted notation octets
        # excludes loopback network 0.0.0.0
        # excludes reserved space >= 224.0.0.0
        # excludes network & broadcast addresses
        # (first & last IP address of each class)
        r"(?P<public_ip>"
        r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
        r"" + ip_middle_octet + "{2}"
        r"" + ip_last_octet + ")"
        r"|"
        # host name
        r"(?:(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)"
        # domain name
        r"(?:\.(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)*"
        # TLD identifier
        r"(?:\.(?:[a-z\u00a1-\uffff]{2,}))"
        r")"
        # port number
        r"(?::\d{2,5})?"
        # resource path
        r"(?:/\S*)?"
        # query string
        r"(?:\?\S*)?"
        r"$",
        re.UNICODE | re.IGNORECASE
    )
    pattern = re.compile(regex)
    if pattern.match(s):
        return True
    return False



from flask import Flask, cli, render_template
from flasgger import Swagger
from recon.core import base
from recon.core.constants import BANNER_WEB
from recon.core.web.db import Tasks
from redis import Redis
import os
import rq

# disable the development server warning banner
cli.show_server_banner = lambda *x: None

print(BANNER_WEB)

# create an application-wide framework and tasks instance
recon = base.Recon(check=False, analytics=False, marketplace=False)
recon.start(base.Mode.WEB)
tasks = Tasks(recon)

# configuration
DEBUG = False
SECRET_KEY = 'we keep no secrets here.'
JSON_SORT_KEYS = False
REDIS_URL = os.environ.get('REDIS_URL', 'redis://')
SWAGGER = {
    'title': 'Swagger',
    'info': {
        'title': 'Recon-API',
        'description': 'A RESTful API for Recon-ng',
        'version': '0.0.1',
    },
    'uiversion': 3,
    'specs_route': '/api/',
}
WORKSPACE = recon.workspace.split('/')[-1]
print((f" * Workspace initialized: {WORKSPACE}"))

def create_app():

    # setting the static_url_path to blank serves static files from the web root
    app = Flask(__name__, static_url_path='')
    app.config.from_object(__name__)

    Swagger(app, template_file='definitions.yaml')

    app.redis = Redis.from_url(app.config['REDIS_URL'])
    app.task_queue = rq.Queue('recon-tasks', connection=app.redis)

    @app.after_request
    def disable_cache(response):
        response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
        return response

    @app.route('/')
    def index():
        return render_template('index.html', workspaces=recon._get_workspaces())

    from recon.core.web.api import resources
    app.register_blueprint(resources)

    return app



import re


class ValidationException(Exception):
    def __init__(self, input, validator):
        Exception.__init__(self, f"Input failed {validator} validation: {input}")


class BaseValidator(object):

    def __init__(self, regex, validator=None, flags=0):
        if isinstance(regex, str):
            self.match_object = re.compile(regex, flags)
        else:
            self.match_object = regex
        self.validator = validator

    def validate(self, value):
        if self.match_object.match(value) is None:
            raise ValidationException(value, self.validator)


class DomainValidator(BaseValidator):

    def __init__(self):
        regex = (
            r"(?=^.{4,253}\.?$)(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}\.?$)"
        )
        super(DomainValidator, self).__init__(regex, 'domain')


class UrlValidator(BaseValidator):

    def __init__(self):
        regex = (
            # {http,ftp}s:// (not required)
            r"^((?:http|ftp)s?://)?"
            # Domain
            r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+"
            r"(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
            # Localhost
            r"localhost|"
            # IPv6 address
            r"\[[a-f0-9:]+\]|"
            # IPv4 address
            r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
            # Optional port
            r"(?::\d+)?"
            # Path
            r"(?:/?|[/?]\S+)$"
        )
        super(UrlValidator, self).__init__(regex, 'url')


class EmailValidator(BaseValidator):

    def __init__(self):
        regex = (
            r"^[a-zA-Z0-9.!#$%&'*+\/=?^_`{|}~-]+@[a-zA-Z0-9]"
            r"(?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9]"
            r"(?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$"
        )
        super(EmailValidator, self).__init__(regex, 'email')



from recon.core import base
from recon.core.web.db import Tasks
from rq import get_current_job
import traceback

# These tasks exist outside the web directory to avoid loading the entire 
# application (which reloads the framework) on every task execution.

def run_module(workspace, module):

    results = {}
    try:
        # instantiate important objects
        job = get_current_job()
        recon = base.Recon(check=False, analytics=False, marketplace=False)
        recon.start(base.Mode.JOB, workspace=workspace)
        tasks = Tasks(recon)
        # update the task's status
        tasks.update_task(job.get_id(), status=job.get_status())
        # execute the task
        module = recon._loaded_modules.get(module)
        module.run()
    except Exception as e:
        results['error'] = {
            'type': str(type(e)),
            'message': str(e),
            'traceback': traceback.format_exc(),
        }
    results['summary'] = module._summary_counts
    # update the task's status and results
    tasks.update_task(job.get_id(), status='finished', result=results)
    return results



import mechanize
import socket

class BrowserMixin(object):

    def get_browser(self):
        '''Returns a mechanize.Browser object configured with the framework's global options.'''
        br = mechanize.Browser()
        # set the user-agent header
        br.addheaders = [('User-agent', self._global_options['user-agent'])]
        # set debug options
        if self._global_options['verbosity'] >= 2:
            br.set_debug_http(True)
            br.set_debug_redirects(True)
            br.set_debug_responses(True)
        # set proxy
        if self._global_options['proxy']:
            br.set_proxies({'http': self._global_options['proxy'], 'https': self._global_options['proxy']})
        # additional settings
        br.set_handle_robots(False)
        # set timeout
        socket.setdefaulttimeout(self._global_options['timeout'])
        return br



from recon.core import framework
import time
import urllib.parse


class TwitterMixin(object):

    def get_twitter_oauth_token(self):
        token_name = 'twitter_token'
        token = self.get_key(token_name)
        if token:
            return token
        twitter_key = self.get_key('twitter_api')
        twitter_secret = self.get_key('twitter_secret')
        url = 'https://api.twitter.com/oauth2/token'
        auth = (twitter_key, twitter_secret)
        headers = {'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8'}
        payload = {'grant_type': 'client_credentials'}
        resp = self.request('POST', url, auth=auth, headers=headers, data=payload)
        if 'errors' in resp.json():
            raise framework.FrameworkException(f"{resp.json()['errors'][0]['message']}, {resp.json()['errors'][0]['label']}")
        access_token = resp.json()['access_token']
        self.add_key(token_name, access_token)
        return access_token

    def search_twitter_api(self, payload, limit=False):
        headers = {'Authorization': f"Bearer {self.get_twitter_oauth_token()}"}
        url = 'https://api.twitter.com/1.1/search/tweets.json'
        results = []
        while True:
            resp = self.request('GET', url, params=payload, headers=headers)
            if limit:
                # app auth rate limit for search/tweets is 450/15min
                time.sleep(2)
            jsonobj = resp.json()
            for item in ['error', 'errors']:
                if item in jsonobj:
                    raise framework.FrameworkException(jsonobj[item])
            results += jsonobj['statuses']
            if 'next_results' in jsonobj['search_metadata']:
                max_id = urllib.parse.parse_qs(jsonobj['search_metadata']['next_results'][1:])['max_id'][0]
                payload['max_id'] = max_id
                continue
            break
        return results



__author__    = 'Tim Tomes (@lanmaster53)'

from datetime import datetime
from pathlib import Path
from urllib.parse import urljoin
import errno
import imp
import json
import os
import random
import re
import shutil
import sys
import yaml
import builtins

# import framework libs
from recon.core import framework
from recon.core.constants import BANNER, BANNER_SMALL

# set the __version__ variable based on the VERSION file
exec(open(os.path.join(Path(os.path.abspath(__file__)).parents[2], 'VERSION')).read())

# using stdout to spool causes tab complete issues
# therefore, override print function
# use a lock for thread safe console and spool output
from threading import Lock
_print_lock = Lock()
# spooling system
def spool_print(*args, **kwargs):
    with _print_lock:
        if framework.Framework._spool:
            framework.Framework._spool.write(f"{args[0]}{os.linesep}")
            framework.Framework._spool.flush()
        # disable terminal output for server jobs
        if framework.Framework._mode == Mode.JOB:
            return
        # new print function must still use the old print function via the backup
        builtins._print(*args, **kwargs)
# make a builtin backup of the original print function
builtins._print = print
# override the builtin print function with the new print function
builtins.print = spool_print

#=================================================
# BASE CLASS
#=================================================

class Recon(framework.Framework):

    repo_url = 'https://raw.githubusercontent.com/lanmaster53/recon-ng-modules/master/'

    def __init__(self, check=True, analytics=True, marketplace=True, accessible=False):
        framework.Framework.__init__(self, 'base')
        self._name = 'recon-ng'
        self._prompt_template = '{}[{}] > '
        self._base_prompt = self._prompt_template.format('', self._name)
        # set toggle flags
        self._check = check
        self._analytics = analytics
        self._marketplace = marketplace
        self._accessible = accessible
        # set path variables
        self.app_path = framework.Framework.app_path = sys.path[0]
        self.core_path = framework.Framework.core_path = os.path.join(self.app_path, 'core')
        self.home_path = framework.Framework.home_path = os.path.join(os.path.expanduser('~'), '.recon-ng')
        self.mod_path = framework.Framework.mod_path = os.path.join(self.home_path, 'modules')
        self.data_path = framework.Framework.data_path = os.path.join(self.home_path, 'data')
        self.spaces_path = framework.Framework.spaces_path = os.path.join(self.home_path, 'workspaces')

    def start(self, mode, workspace='default'):
        # initialize framework components
        self._mode = framework.Framework._mode = mode
        self._init_global_options()
        self._init_home()
        self._init_workspace(workspace)
        self._check_version()
        if self._mode == Mode.CONSOLE:
            self._print_banner()
            self.cmdloop()

    #==================================================
    # SUPPORT METHODS
    #==================================================

    def _init_global_options(self):
        self.options = self._global_options
        self.register_option('nameserver', '8.8.8.8', True, 'default nameserver for the resolver mixin')
        self.register_option('proxy', None, False, 'proxy server (address:port)')
        self.register_option('threads', 10, True, 'number of threads (where applicable)')
        self.register_option('timeout', 10, True, 'socket timeout (seconds)')
        self.register_option('user-agent', f"Recon-ng/v{__version__.split('.')[0]}", True, 'user-agent string')
        self.register_option('verbosity', 1, True, 'verbosity level (0 = minimal, 1 = verbose, 2 = debug)')

    def _init_home(self):
        # initialize home folder
        if not os.path.exists(self.home_path):
            os.makedirs(self.home_path)
        # initialize keys database
        self._query_keys('CREATE TABLE IF NOT EXISTS keys (name TEXT PRIMARY KEY, value TEXT)')
        # initialize module index
        self._fetch_module_index()

    def _check_version(self):
        if self._check:
            pattern = r"'(\d+\.\d+\.\d+[^']*)'"
            remote = 0
            local = 0
            try:
                remote = re.search(pattern, self.request('GET', 'https://raw.githubusercontent.com/lanmaster53/recon-ng/master/VERSION').text).group(1)
                local = re.search(pattern, open('VERSION').read()).group(1)
            except Exception as e:
                self.error(f"Version check failed ({type(e).__name__}).")
                #self.print_exception()
            if remote != local:
                self.alert('Your version of Recon-ng does not match the latest release.')
                self.alert('Please consider updating before further use.')
                self.output(f"Remote version:  {remote}")
                self.output(f"Local version:   {local}")
        else:
            self.alert('Version check disabled.')

    def _print_banner(self):
        banner = BANNER
        banner_len = len(max(banner.split(os.linesep), key=len))
        author = '{0:^{1}}'.format(f"{framework.Colors.O}[{self._name} v{__version__}, {__author__}]{framework.Colors.N}", banner_len + 8)
        if self._accessible:
            banner = BANNER_SMALL
            author = f"{framework.Colors.O}{self._name}, version {__version__}, by {__author__}{framework.Colors.N}"
        print(banner)
        print(author)
        print('')
        counts = [(len(self._loaded_category[x]), x) for x in self._loaded_category]
        if counts:
            count_len = len(max([self.to_unicode_str(x[0]) for x in counts], key=len))
            for count in sorted(counts, reverse=True):
                cnt = f"[{count[0]}]"
                print(f"{framework.Colors.B}{cnt.ljust(count_len+2)} {count[1].title()} modules{framework.Colors.N}")
                # create dynamic easter egg command based on counts
                setattr(self, f"do_{count[0]}", self._menu_egg)
        else:
            self.alert('No modules enabled/installed.')
        print('')

    def _send_analytics(self, cd):
        if self._analytics:
            try:
                cid_path = os.path.join(self.home_path, '.cid')
                if not os.path.exists(cid_path):
                    # create the cid and file
                    import uuid
                    with open(cid_path, 'w') as fp:
                        fp.write(self.to_unicode_str(uuid.uuid4()))
                with open(cid_path) as fp:
                    cid = fp.read().strip()
                params = {
                        'v': 1,
                        'tid': 'UA-52269615-2',
                        'cid': cid,
                        't': 'screenview',
                        'an': 'Recon-ng',
                        'av': __version__,
                        'cd': cd
                        }
                self.request('GET', 'https://www.google-analytics.com/collect', params=params)
            except Exception as e:
                self.debug(f"Analytics failed ({type(e).__name__}).")
                #self.print_exception()
                return
        else:
            self.debug('Analytics disabled.')

    def _menu_egg(self, params):
        eggs = [
            'Really? A menu option? Try again.',
            'You clearly need \'help\'.',
            'That makes no sense to me.',
            '*grunt* *grunt* Nope. I got nothin\'.',
            'Wait for it...',
            'This is not the Social Engineering Toolkit.',
            'Don\'t you think if that worked the numbers would at least be in order?',
            'Reserving that option for the next-NEXT generation of the framework.',
            'You\'ve clearly got the wrong framework. Attempting to start SET...',
            '1980 called. They want there menu driven UI back.',
        ]
        print(random.choice(eggs))
        return

    #==================================================
    # WORKSPACE METHODS
    #==================================================

    def _init_workspace(self, workspace):
        if not workspace:
            return
        path = os.path.join(self.spaces_path, workspace)
        self.workspace = framework.Framework.workspace = path
        if not os.path.exists(path):
            os.makedirs(path)
            self._create_db()
        else:
            self._migrate_db()
        # set workspace prompt
        self.prompt = self._prompt_template.format(self._base_prompt[:-3], self.workspace.split('/')[-1])
        # load workspace configuration
        self._load_config()
        # reload modules after config to populate options
        self._load_modules()
        return True

    def remove_workspace(self, workspace):
        path = os.path.join(self.spaces_path, workspace)
        try:
            shutil.rmtree(path)
        except OSError:
            return False
        if workspace == self.workspace.split('/')[-1]:
            self._init_workspace('default')
        return True

    def _get_workspaces(self):
        workspaces = []
        path = self.spaces_path
        for name in os.listdir(path):
            if os.path.isdir(os.path.join(path, name)):
                workspaces.append(name)
        return workspaces

    def _get_snapshots(self):
        snapshots = []
        for f in os.listdir(self.workspace):
            if re.search(r'^snapshot_\d{14}.db$', f):
                snapshots.append(f)
        return snapshots

    def _create_db(self):
        self.query('CREATE TABLE IF NOT EXISTS domains (domain TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS companies (company TEXT, description TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS netblocks (netblock TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS locations (latitude TEXT, longitude TEXT, street_address TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS vulnerabilities (host TEXT, reference TEXT, example TEXT, publish_date TEXT, category TEXT, status TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS ports (ip_address TEXT, host TEXT, port TEXT, protocol TEXT, banner TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS hosts (host TEXT, ip_address TEXT, region TEXT, country TEXT, latitude TEXT, longitude TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS contacts (first_name TEXT, middle_name TEXT, last_name TEXT, email TEXT, title TEXT, region TEXT, country TEXT, phone TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS credentials (username TEXT, password TEXT, hash TEXT, type TEXT, leak TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS leaks (leak_id TEXT, description TEXT, source_refs TEXT, leak_type TEXT, title TEXT, import_date TEXT, leak_date TEXT, attackers TEXT, num_entries TEXT, score TEXT, num_domains_affected TEXT, attack_method TEXT, target_industries TEXT, password_hash TEXT, password_type TEXT, targets TEXT, media_refs TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS pushpins (source TEXT, screen_name TEXT, profile_name TEXT, profile_url TEXT, media_url TEXT, thumb_url TEXT, message TEXT, latitude TEXT, longitude TEXT, time TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS profiles (username TEXT, resource TEXT, url TEXT, category TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS repositories (name TEXT, owner TEXT, description TEXT, resource TEXT, category TEXT, url TEXT, notes TEXT, module TEXT)')
        self.query('CREATE TABLE IF NOT EXISTS dashboard (module TEXT PRIMARY KEY, runs INT)')
        self.query('PRAGMA user_version = 10')

    def _migrate_db(self):
        db_version = lambda self: self.query('PRAGMA user_version')[0][0]
        db_orig = db_version(self)
        if db_version(self) == 0:
            # add mname column to contacts table
            tmp = self.get_random_str(20)
            self.query(f"ALTER TABLE contacts RENAME TO {tmp}")
            self.query('CREATE TABLE contacts (fname TEXT, mname TEXT, lname TEXT, email TEXT, title TEXT, region TEXT, country TEXT)')
            self.query(f"INSERT INTO contacts (fname, lname, email, title, region, country) SELECT fname, lname, email, title, region, country FROM {tmp}")
            self.query(f"DROP TABLE {tmp}")
            self.query('PRAGMA user_version = 1')
        if db_version(self) == 1:
            # rename name columns
            tmp = self.get_random_str(20)
            self.query(f"ALTER TABLE contacts RENAME TO {tmp}")
            self.query('CREATE TABLE contacts (first_name TEXT, middle_name TEXT, last_name TEXT, email TEXT, title TEXT, region TEXT, country TEXT)')
            self.query(f"INSERT INTO contacts (first_name, middle_name, last_name, email, title, region, country) SELECT fname, mname, lname, email, title, region, country FROM {tmp}")
            self.query(f"DROP TABLE {tmp}")
            # rename pushpin table
            self.query('ALTER TABLE pushpin RENAME TO pushpins')
            # add new tables
            self.query('CREATE TABLE IF NOT EXISTS domains (domain TEXT)')
            self.query('CREATE TABLE IF NOT EXISTS companies (company TEXT, description TEXT)')
            self.query('CREATE TABLE IF NOT EXISTS netblocks (netblock TEXT)')
            self.query('CREATE TABLE IF NOT EXISTS locations (latitude TEXT, longitude TEXT)')
            self.query('CREATE TABLE IF NOT EXISTS vulnerabilities (host TEXT, reference TEXT, example TEXT, publish_date TEXT, category TEXT)')
            self.query('CREATE TABLE IF NOT EXISTS ports (ip_address TEXT, host TEXT, port TEXT, protocol TEXT)')
            self.query('CREATE TABLE IF NOT EXISTS leaks (leak_id TEXT, description TEXT, source_refs TEXT, leak_type TEXT, title TEXT, import_date TEXT, leak_date TEXT, attackers TEXT, num_entries TEXT, score TEXT, num_domains_affected TEXT, attack_method TEXT, target_industries TEXT, password_hash TEXT, targets TEXT, media_refs TEXT)')
            self.query('PRAGMA user_version = 2')
        if db_version(self) == 2:
            # add street_address column to locations table
            self.query('ALTER TABLE locations ADD COLUMN street_address TEXT')
            self.query('PRAGMA user_version = 3')
        if db_version(self) == 3:
            # account for db_version bug
            if 'creds' in self.get_tables():
                # rename creds table
                self.query('ALTER TABLE creds RENAME TO credentials')
            self.query('PRAGMA user_version = 4')
        if db_version(self) == 4:
            # add status column to vulnerabilities table
            if 'status' not in [x[0] for x in self.get_columns('vulnerabilities')]:
                self.query('ALTER TABLE vulnerabilities ADD COLUMN status TEXT')
            # add module column to all tables
            for table in ['domains', 'companies', 'netblocks', 'locations', 'vulnerabilities', 'ports', 'hosts', 'contacts', 'credentials', 'leaks', 'pushpins']:
                if 'module' not in [x[0] for x in self.get_columns(table)]:
                    self.query(f"ALTER TABLE {table} ADD COLUMN module TEXT")
            self.query('PRAGMA user_version = 5')
        if db_version(self) == 5:
            # add profile table
            self.query('CREATE TABLE IF NOT EXISTS profiles (username TEXT, resource TEXT, url TEXT, category TEXT, notes TEXT, module TEXT)')
            self.query('PRAGMA user_version = 6')
        if db_version(self) == 6:
            # add repositories table
            self.query('CREATE TABLE IF NOT EXISTS repositories (name TEXT, owner TEXT, description TEXT, resource TEXT, category TEXT, url TEXT, module TEXT)')
            self.query('PRAGMA user_version = 7')
        if db_version(self) == 7:
            # add password_type column to leaks table
            self.query('ALTER TABLE leaks ADD COLUMN password_type TEXT')
            self.query('UPDATE leaks SET password_type=\'unknown\'')
            self.query('PRAGMA user_version = 8')
        if db_version(self) == 8:
            # add banner column to ports table
            self.query('ALTER TABLE ports ADD COLUMN banner TEXT')
            # add notes column to all tables
            for table in ['domains', 'companies', 'netblocks', 'locations', 'vulnerabilities', 'ports', 'hosts', 'contacts', 'credentials', 'leaks', 'pushpins', 'profiles', 'repositories']:
                if 'notes' not in [x[0] for x in self.get_columns(table)]:
                    self.query(f"ALTER TABLE {table} ADD COLUMN notes TEXT")
            self.query('PRAGMA user_version = 9')
        if db_version(self) == 9:
            # add phone column to contacts table
            self.query('ALTER TABLE contacts ADD COLUMN phone TEXT')
            self.query('PRAGMA user_version = 10')
        if db_orig != db_version(self):
            self.alert(f"Database upgraded to version {db_version(self)}.")

    #==================================================
    # MODULE METHODS
    #==================================================

    def _request_file_from_repo(self, path):
        resp = self.request('GET', urljoin(self.repo_url, path))
        if resp.status_code != 200:
            raise framework.FrameworkException(f"Invalid response from module repository ({resp.status_code}).")
        return resp

    def _write_local_file(self, path, content):
        dirpath = os.path.sep.join(path.split(os.path.sep)[:-1])
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        with open(path, 'w') as outfile:
            outfile.write(content)

    def _remove_empty_dirs(self, base_path):
        for root, dirs, files in os.walk(base_path, topdown=False):
            for rel_path in dirs:
                abs_path = os.path.join(root, rel_path)
                if os.path.exists(abs_path):
                    if not os.listdir(abs_path):
                        os.removedirs(abs_path)

    def _fetch_module_index(self):
        if self._marketplace:
            content = '[]'
            self.debug('Fetching index file...')
            try:
                resp = self._request_file_from_repo('modules.yml')
            except Exception as e:
                self.error(f"Unable to synchronize module index. ({type(e).__name__})")
                #self.print_exception()
                return
            content = resp.text
            path = os.path.join(self.home_path, 'modules.yml')
            self._write_local_file(path, content)
        else:
            self.alert('Marketplace disabled.')

    def _update_module_index(self):
        self.debug('Updating index file...')
        # initialize module index
        self._module_index = []
        # load module index from local copy
        path = os.path.join(self.home_path, 'modules.yml')
        if os.path.exists(path):
            with open(path, 'r') as infile:
                self._module_index = yaml.safe_load(infile)
            # add status to index for each module
            for module in self._module_index:
                status = 'not installed'
                if module['path'] in self._loaded_category.get('disabled', []):
                    status = 'disabled'
                elif module['path'] in self._loaded_modules.keys():
                    status = 'installed'
                    loaded = self._loaded_modules[module['path']]
                    if loaded.meta['version'] != module['version']:
                        status = 'outdated'
                module['status'] = status

    def _search_module_index(self, s):
        keys = ('path', 'name', 'description', 'status')
        modules = []
        for module in self._module_index:
            for key in keys:
                if re.search(s, module[key]):
                    modules.append(module)
                    break
        return modules

    def _get_module_from_index(self, path):
        for module in self._module_index:
            if module['path'] == path:
                return module
        return None

    def _install_module(self, path):
        # download supporting data files
        downloads = {}
        files = self._get_module_from_index(path).get('files', [])
        for filename in files:
            try:
                resp = self._request_file_from_repo('/'.join(['data', filename]))
            except:
                self.error(f"Supporting file download for {path} failed: ({filename})")
                self.error('Module installation aborted.')
                raise
            abs_path = os.path.join(self.data_path, filename)
            downloads[abs_path] = resp.text
        # download the module
        rel_path = '.'.join([path, 'py'])
        try:
            resp = self._request_file_from_repo('/'.join(['modules', rel_path]))
        except:
            self.error(f"Module installation failed: {path}")
            raise
        abs_path = os.path.join(self.mod_path, rel_path)
        downloads[abs_path] = resp.text
        # install the module
        for abs_path, content in downloads.items():
            self._write_local_file(abs_path, content)
        self.output(f"Module installed: {path}")

    def _remove_module(self, path):
        # remove the module
        rel_path = '.'.join([path, 'py'])
        abs_path = os.path.join(self.mod_path, rel_path)
        os.remove(abs_path)
        # remove supporting data files
        files = self._get_module_from_index(path).get('files', [])
        for filename in files:
            abs_path = os.path.join(self.data_path, filename)
            if os.path.exists(abs_path):
                os.remove(abs_path)
        self.output(f"Module removed: {path}")

    def _load_modules(self):
        self._loaded_category = {}
        self._loaded_modules = framework.Framework._loaded_modules = {}
        # crawl the module directory and build the module tree
        for dirpath, dirnames, filenames in os.walk(self.mod_path, followlinks=True):
            # remove hidden files and directories
            filenames = [f for f in filenames if not f[0] == '.']
            dirnames[:] = [d for d in dirnames if not d[0] == '.']
            if len(filenames) > 0:
                for filename in [f for f in filenames if f.endswith('.py')]:
                    self._load_module(dirpath, filename)
        # cleanup module directory
        self._remove_empty_dirs(self.mod_path)
        # update module index
        self._update_module_index()

    def _load_module(self, dirpath, filename):
        mod_name = filename.split('.')[0]
        mod_category = re.search('/modules/([^/]*)', dirpath).group(1)
        mod_dispname = '/'.join(re.split('/modules/', dirpath)[-1].split('/') + [mod_name])
        mod_loadname = mod_dispname.replace('/', '_')
        mod_loadpath = os.path.join(dirpath, filename)
        mod_file = open(mod_loadpath)
        try:
            # import the module into memory
            mod = imp.load_source(mod_loadname, mod_loadpath, mod_file)
            __import__(mod_loadname)
            # add the module to the framework's loaded modules
            self._loaded_modules[mod_dispname] = sys.modules[mod_loadname].Module(mod_dispname)
            self._categorize_module(mod_category, mod_dispname)
            # return indication of success to support module reload
            return True
        except ImportError as e:
            # notify the user of missing dependencies
            self.error(f"Module '{mod_dispname}' disabled. Dependency required: '{self.to_unicode_str(e)[16:]}'")
        except:
            # notify the user of errors
            self.print_exception()
            self.error(f"Module '{mod_dispname}' disabled.")
        # remove the module from the framework's loaded modules
        self._loaded_modules.pop(mod_dispname, None)
        self._categorize_module('disabled', mod_dispname)

    def _categorize_module(self, category, module):
        if not category in self._loaded_category:
            self._loaded_category[category] = []
        self._loaded_category[category].append(module)

    #==================================================
    # COMMAND METHODS
    #==================================================

    def do_index(self, params):
        '''Creates a module index (dev only)'''
        mod_path, file_name = self._parse_params(params)
        if not mod_path:
            self.help_index()
            return
        self.output('Building index markup...')
        yaml_objs = []
        modules = [m for m in self._loaded_modules.items() if mod_path in m[0] or mod_path == 'all']
        for path, module in sorted(modules, key=lambda k: k[0]):
            yaml_obj = {}
            # not in meta
            yaml_obj['path'] = path
            yaml_obj['last_updated'] = datetime.strftime(datetime.now(), '%Y-%m-%d')
            # meta required
            yaml_obj['author'] = module.meta.get('author')
            yaml_obj['name'] = module.meta.get('name')
            yaml_obj['description'] = module.meta.get('description')
            yaml_obj['version'] = module.meta.get('version', '1.0')
            # meta optional
            #yaml_obj['comments'] = module.meta.get('comments', [])
            yaml_obj['dependencies'] = module.meta.get('dependencies', [])
            yaml_obj['files'] = module.meta.get('files', [])
            #yaml_obj['options'] = module.meta.get('options', [])
            #yaml_obj['query'] = module.meta.get('query', '')
            yaml_obj['required_keys'] = module.meta.get('required_keys', [])
            yaml_objs.append(yaml_obj)
        if yaml_objs:
            markup = yaml.safe_dump(yaml_objs)
            print(markup)
            # write to file if index name provided
            if file_name:
                with open(file_name, 'w') as outfile:
                    outfile.write(markup)
                self.output('Module index created.')
        else:
            self.output('No modules found.')

    def do_marketplace(self, params):
        '''Interfaces with the module marketplace'''
        if not self._marketplace:
            self.alert('Marketplace disabled.')
            return
        if not params:
            self.help_marketplace()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('marketplace'):
            return getattr(self, '_do_marketplace_'+arg)(params)
        else:
            self.help_marketplace()

    def _do_marketplace_refresh(self, params):
        '''Refreshes the marketplace index'''
        self._fetch_module_index()
        self._update_module_index()
        self.output('Marketplace index refreshed.')

    def _do_marketplace_search(self, params):
        '''Searches marketplace modules'''
        modules = [m for m in self._module_index]
        if params:
            self.output(f"Searching module index for '{params}'...")
            modules = self._search_module_index(params)
        if modules:
            rows = []
            for module in sorted(modules, key=lambda m: m['path']):
                row = []
                for key in ('path', 'version', 'status', 'last_updated'):
                    row.append(module[key])
                row.append('*' if module['dependencies'] else '')
                row.append('*' if module['required_keys'] else '')
                rows.append(row)
            header = ('Path', 'Version', 'Status', 'Updated', 'D', 'K')
            self.table(rows, header=header)
            print(f"{self.spacer}D = Has dependencies. See info for details.")
            print(f"{self.spacer}K = Requires keys. See info for details.{os.linesep}")
        else:
            self.error('No modules found.')
            self._help_marketplace_search()

    def _do_marketplace_info(self, params):
        '''Shows detailed information about available modules'''
        if not params:
            self._help_marketplace_info()
            return
        modules = [m for m in self._module_index if params in m['path'] or params == 'all']
        if modules:
            for module in modules:
                rows = []
                for key in ('path', 'name', 'author', 'version', 'last_updated', 'description', 'required_keys', 'dependencies', 'files', 'status'):
                    row = (key, module[key])
                    rows.append(row)
                self.table(rows)
        else:
            self.error('Invalid module path.')

    def _do_marketplace_install(self, params):
        '''Installs modules from the marketplace'''
        if not params:
            self._help_marketplace_install()
            return
        modules = [m for m in self._module_index if params in m['path'] or params == 'all']
        if modules:
            for module in modules:
                self._install_module(module['path'])
            self._do_modules_reload('')
        else:
            self.error('Invalid module path.')

    def _do_marketplace_remove(self, params):
        '''Removes marketplace modules from the framework'''
        if not params:
            self._help_marketplace_remove()
            return
        modules = [m for m in self._module_index if m['status'] in ('installed', 'disabled') and (params in m['path'] or params == 'all')]
        if modules:
            for module in modules:
                self._remove_module(module['path'])
            self._do_modules_reload('')
        else:
            self.error('Invalid module path.')

    def do_workspaces(self, params):
        '''Manages workspaces'''
        if not params:
            self.help_workspaces()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('workspaces'):
            return getattr(self, '_do_workspaces_'+arg)(params)
        else:
            self.help_workspaces()

    def _do_workspaces_list(self, params):
        '''Lists existing workspaces'''
        rows = []
        for workspace in self._get_workspaces():
            db_path = os.path.join(self.spaces_path, workspace, 'data.db')
            modified = datetime.fromtimestamp(os.path.getmtime(db_path)).strftime('%Y-%m-%d %H:%M:%S')
            rows.append((workspace, modified))
        rows.sort(key=lambda x: x[0])
        self.table(rows, header=['Workspaces', 'Modified'])

    def _do_workspaces_create(self, params):
        '''Creates a new workspace'''
        if not params:
            self._help_workspaces_create()
            return
        if not self._init_workspace(params):
            self.output(f"Unable to create '{params}' workspace.")

    def _do_workspaces_load(self, params):
        '''Loads an existing workspace'''
        if not params:
            self._help_workspaces_load()
            return
        if params in self._get_workspaces():
            if not self._init_workspace(params):
                self.output(f"Unable to initialize '{params}' workspace.")
        else:
            self.output('Invalid workspace name.')

    def _do_workspaces_remove(self, params):
        '''Removes an existing workspace'''
        if not params:
            self._help_workspaces_remove()
            return
        if not self.remove_workspace(params):
            self.output(f"Unable to remove '{params}' workspace.")

    def do_snapshots(self, params):
        '''Manages workspace snapshots'''
        if not params:
            self.help_snapshots()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('snapshots'):
            return getattr(self, '_do_snapshots_'+arg)(params)
        else:
            self.help_snapshots()

    def _do_snapshots_list(self, params):
        '''Lists existing database snapshots'''
        snapshots = self._get_snapshots()
        if snapshots:
            self.table([[x] for x in snapshots], header=['Snapshots'])
        else:
            self.output('This workspace has no snapshots.')

    def _do_snapshots_take(self, params):
        '''Takes a snapshot of the current database'''
        ts = datetime.strftime(datetime.now(), '%Y%m%d%H%M%S')
        snapshot = f"snapshot_{ts}.db"
        src = os.path.join(self.workspace, 'data.db')
        dst = os.path.join(self.workspace, snapshot)
        shutil.copyfile(src, dst)
        self.output(f"Snapshot created: {snapshot}")

    def _do_snapshots_load(self, params):
        '''Loads an existing database snapshot'''
        if not params:
            self._help_snapshots_load()
            return
        if params in self._get_snapshots():
            src = os.path.join(self.workspace, params)
            dst = os.path.join(self.workspace, 'data.db')
            shutil.copyfile(src, dst)
            self.output(f"Snapshot loaded: {params}")
        else:
            self.error(f"No snapshot named '{params}'.")

    def _do_snapshots_remove(self, params):
        '''Removes an existing snapshot'''
        if not params:
            self._help_snapshots_remove()
            return
        if params in self._get_snapshots():
            os.remove(os.path.join(self.workspace, params))
            self.output(f"Snapshot removed: {params}")
        else:
            self.error(f"No snapshot named '{params}'.")

    def _do_modules_load(self, params):
        '''Loads a module'''
        # validate global options before loading the module
        try:
            self._validate_options()
        except framework.FrameworkException as e:
            self.error(e)
            return
        if not params:
            self._help_modules_load()
            return
        # finds any modules that contain params
        modules = self._match_modules(params)
        # notify the user if none or multiple modules are found
        if len(modules) != 1:
            if not modules:
                self.error('Invalid module name.')
            else:
                self.output(f"Multiple modules match '{params}'.")
                self._list_modules(modules)
            return
        # load the module
        mod_dispname = modules[0]
        # loop to support reload logic
        while True:
            y = self._loaded_modules[mod_dispname]
            # send analytics information
            mod_loadpath = os.path.abspath(sys.modules[y.__module__].__file__)
            self._send_analytics(mod_dispname)
            # return the loaded module if not in console mode
            if self._mode != Mode.CONSOLE:
                return y
            # begin a command loop
            y.prompt = self._prompt_template.format(self.prompt[:-3], mod_dispname.split('/')[-1])
            try:
                y.cmdloop()
            except KeyboardInterrupt:
                print('')
            if y._exit == 1:
                return True
            if y._reload == 1:
                self.output('Reloading module...')
                # reload the module in memory
                is_loaded = self._load_module(os.path.dirname(mod_loadpath), os.path.basename(mod_loadpath))
                if is_loaded:
                    # reload the module in the framework
                    continue
                # shuffle category counts?
            break

    def _do_modules_reload(self, params):
        '''Reloads installed modules'''
        self.output('Reloading modules...')
        self._load_modules()

    #==================================================
    # HELP METHODS
    #==================================================

    def help_index(self):
        print(getattr(self, 'do_index').__doc__)
        print(f"{os.linesep}Usage: index <module|all> <index>{os.linesep}")

    def help_marketplace(self):
        print(getattr(self, 'do_marketplace').__doc__)
        print(f"{os.linesep}Usage: marketplace <{'|'.join(self._parse_subcommands('marketplace'))}> [...]{os.linesep}")

    def _help_marketplace_search(self):
        print(getattr(self, '_do_marketplace_search').__doc__)
        print(f"{os.linesep}Usage: marketplace search [<regex>]{os.linesep}")

    def _help_marketplace_info(self):
        print(getattr(self, '_do_marketplace_info').__doc__)
        print(f"{os.linesep}Usage: marketplace info <<path>|<prefix>|all>{os.linesep}")

    def _help_marketplace_install(self):
        print(getattr(self, '_do_marketplace_install').__doc__)
        print(f"{os.linesep}Usage: marketplace install <<path>|<prefix>|all>{os.linesep}")

    def _help_marketplace_remove(self):
        print(getattr(self, '_do_marketplace_remove').__doc__)
        print(f"{os.linesep}Usage: marketplace remove <<path>|<prefix>|all>{os.linesep}")

    def help_workspaces(self):
        print(getattr(self, 'do_workspaces').__doc__)
        print(f"{os.linesep}Usage: workspaces <{'|'.join(self._parse_subcommands('workspaces'))}> [...]{os.linesep}")

    def _help_workspaces_create(self):
        print(getattr(self, '_do_workspaces_create').__doc__)
        print(f"{os.linesep}Usage: workspace create <name>{os.linesep}")

    def _help_workspaces_load(self):
        print(getattr(self, '_do_workspaces_load').__doc__)
        print(f"{os.linesep}Usage: workspace load <name>{os.linesep}")

    def _help_workspaces_remove(self):
        print(getattr(self, '_do_workspaces_remove').__doc__)
        print(f"{os.linesep}Usage: workspace remove <name>{os.linesep}")

    def help_snapshots(self):
        print(getattr(self, 'do_snapshots').__doc__)
        print(f"{os.linesep}Usage: snapshots <{'|'.join(self._parse_subcommands('snapshots'))}> [...]{os.linesep}")

    def _help_snapshots_load(self):
        print(getattr(self, '_do_snapshots_load').__doc__)
        print(f"{os.linesep}Usage: snapshots load <name>{os.linesep}")

    def _help_snapshots_remove(self):
        print(getattr(self, '_do_snapshots_remove').__doc__)
        print(f"{os.linesep}Usage: snapshots remove <name>{os.linesep}")

    #==================================================
    # COMPLETE METHODS
    #==================================================

    def complete_index(self, text, line, *ignored):
        if len(line.split(' ')) == 2:
            return [x for x in self._loaded_modules if x.startswith(text)]
        return []

    def complete_marketplace(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('marketplace')
        if arg in subs:
            return getattr(self, '_complete_marketplace_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_marketplace_refresh(self, text, *ignored):
        return []
    _complete_marketplace_search = _complete_marketplace_refresh

    def _complete_marketplace_info(self, text, *ignored):
        return [x['path'] for x in self._module_index if x['path'].startswith(text)]
    _complete_marketplace_install = _complete_marketplace_info

    def _complete_marketplace_remove(self, text, *ignored):
        return [x['path'] for x in self._module_index if x['status'] == 'installed' and x['path'].startswith(text)]

    def complete_workspaces(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('workspaces')
        if arg in subs:
            return getattr(self, '_complete_workspaces_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_workspaces_list(self, text, *ignored):
        return []
    _complete_workspaces_create = _complete_workspaces_list

    def _complete_workspaces_load(self, text, *ignored):
        return [x for x in self._get_workspaces() if x.startswith(text)]
    _complete_workspaces_remove = _complete_workspaces_load

    def complete_snapshots(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('snapshots')
        if arg in subs:
            return getattr(self, '_complete_snapshots_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_snapshots_list(self, text, *ignored):
        return []
    _complete_snapshots_take = _complete_snapshots_list

    def _complete_snapshots_load(self, text, *ignored):
        return [x for x in self._get_snapshots() if x.startswith(text)]
    _complete_snapshots_remove = _complete_snapshots_load

    def _complete_modules_reload(self, text, *ignored):
        return []

#=================================================
# SUPPORT CLASSES
#=================================================

class Mode(object):
   '''Contains constants that represent the state of the interpreter.'''
   CONSOLE = 0
   CLI     = 1
   WEB     = 2
   JOB     = 3
   
   def __init__(self):
       raise NotImplementedError('This class should never be instantiated.')



from lxml.html import fromstring
from http.cookiejar import CookieJar
from recon.core import framework
import os
import re
import time
import urllib.parse
import webbrowser


class GoogleWebMixin(object):

    cookiejar = CookieJar()
    user_agent = 'Lynx/2.8.8dev.3 libwww-FM/2.14 SSL-MM/1.4.1'

    def search_google_web(self, query, limit=0, start_page=1):
        # parsing logic based on https://github.com/maurosoria/s3arch
        url = 'https://www.google.com/search'
        num = 100
        page = start_page
        set_page = lambda x: (x - 1) * num
        payload = {'q':query, 'start':set_page(page), 'num':num, 'complete':0}
        headers = {'user-agent': self.user_agent}
        results = []
        self.verbose(f"Searching Google for: {query}")
        while True:
            resp = self.request('GET', url, params=payload, headers=headers, allow_redirects=False, cookies=self.cookiejar)
            # detect and handle captchas
            # results = 200, first visit = 302, actual captcha = 503
            count = 0
            if resp.status_code != 200:
                if resp.status_code == 302:
                    self.error('Google CAPTCHA triggered. No bypass available.')
                else:
                    self.error('Google encountered an unknown error.')
                break
            tree = fromstring(resp.text)
            links = tree.xpath('//a/@href')
            regmatch = re.compile(r'^/url\?q=[^/]')
            for link in links:
                if regmatch.match(link) != None and 'http://webcache.googleusercontent.com' not in link:
                    results.append(urllib.parse.unquote_plus(link[7:link.find('&')]))
            # check limit
            if limit == page:
                break
            page += 1
            payload['start'] = set_page(page)
            # check for more pages
            if '>Next</' not in resp.text:
                break
        return results


class GoogleAPIMixin(object):

    def search_google_api(self, query, limit=0):
        api_key = self.get_key('google_api')
        cse_id = self.get_key('google_cse')
        url = 'https://www.googleapis.com/customsearch/v1'
        payload = {'alt': 'json', 'prettyPrint': 'false', 'key': api_key, 'cx': cse_id, 'q': query}
        results = []
        cnt = 0
        self.verbose(f"Searching Google API for: {query}")
        while True:
            resp = self.request('GET', url, params=payload)
            if resp.json() == None:
                raise framework.FrameworkException(f"Invalid JSON response.{os.linesep}{resp.text}")
            # add new results
            if 'items' in resp.json():
                results.extend(resp.json()['items'])
            # increment and check the limit
            cnt += 1
            if limit == cnt:
                break
            # check for more pages
            if not 'nextPage' in resp.json()['queries']:
                break
            payload['start'] = resp.json()['queries']['nextPage'][0]['startIndex']
        return results


class BingAPIMixin(object):

    def search_bing_api(self, query, limit=0):
        url = 'https://api.bing.microsoft.com/v7.0/search'
        payload = {'q': query, 'count': 50, 'offset': 0, 'responseFilter': 'WebPages'}
        headers = {'Ocp-Apim-Subscription-Key': self.get_key('bing_api')}
        results = []
        cnt = 0
        self.verbose(f"Searching Bing API for: {query}")
        while True:
            resp = self.request('GET', url, params=payload, headers=headers)
            if resp.json() == None:
                raise framework.FrameworkException(f"Invalid JSON response.{os.linesep}{resp.text}")
            #elif 'error' in resp.json():
            elif resp.status_code == 401:
                error = resp.json()['error']
                raise framework.FrameworkException(f"({error['code']}) {error['message']}")
            # add new results, or if there's no more, return what we have...
            if 'webPages' in resp.json():
                results.extend(resp.json()['webPages']['value'])
            else:
                return results
            # increment and check the limit
            cnt += 1
            if limit == cnt:
                break
            # check for more pages
            # https://msdn.microsoft.com/en-us/library/dn760787.aspx
            if payload['offset'] > (resp.json()['webPages']['totalEstimatedMatches'] - payload['count']):
                break
            # set the payload for the next request
            payload['offset'] += payload['count']
        return results


class ShodanAPIMixin(object):

    def search_shodan_api(self, query, limit=0):
        api_key = self.get_key('shodan_api')
        url = 'https://api.shodan.io/shodan/host/search'
        payload = {'query': query, 'key': api_key}
        results = []
        cnt = 0
        page = 1
        self.verbose(f"Searching Shodan API for: {query}")
        while True:
            time.sleep(1)
            resp = self.request('GET', url, params=payload)
            if resp.json() == None:
                raise framework.FrameworkException(f"Invalid JSON response.{os.linesep}{resp.text}")
            if 'error' in resp.json():
                raise framework.FrameworkException(resp.json()['error'])
            if not resp.json()['matches']:
                break
            # add new results
            results.extend(resp.json()['matches'])
            # increment and check the limit
            cnt += 1
            if limit == cnt:
                break
            # next page
            page += 1
            payload['page'] = page
        return results



from urllib.parse import urlparse
import html
import re

def parse_hostname(s):
    host = urlparse(s)
    if not host.scheme:
        host = urlparse('//'+s)
    return host.netloc

def parse_emails(s):
    return re.findall(r'([^\s]+@[^\s]+)', s)

def parse_name(s):
    elements = [html.unescape(x) for x in s.strip().split()]
    # remove prefixes and suffixes
    names = []
    for i in range(0,len(elements)):
        # preserve initials
        if re.search(r'^\w\.$', elements[i]):
            elements[i] = elements[i][:-1]
        # remove unecessary prefixes and suffixes
        elif re.search(r'(?:\.|^the$|^jr$|^sr$|^I{2,3}$)', elements[i], re.IGNORECASE):
            continue
        names.append(elements[i])
    # make sense of the remaining elements
    if len(names) > 3:
        names[2:] = [' '.join(names[2:])]
    # clean up any remaining garbage characters
    names = [re.sub(r"[,']", '', x) for x in names]
    # set values and return names
    fname = names[0] if len(names) >= 1 else None
    mname = names[1] if len(names) >= 3 else None
    lname = names[-1] if len(names) >= 2 else None
    return fname, mname, lname



from flask import Blueprint, current_app, request, abort
from flask_restful import Resource, Api
from recon.core.web import recon, tasks
from recon.core.web.utils import columnize
from recon.core.web.constants import EXPORTS, REPORTS

resources = Blueprint('resources', __name__, url_prefix='/api')
api = Api()
api.init_app(resources)


class TaskList(Resource):

    def get(self):
        '''
        Gets all tasks for the current workspace
        ---
        responses:
            200:
                description: List of tasks
                schema:
                    properties:
                        tasks:
                            type: array
                            items:
                                $ref: '#/definitions/Task'
                    required:
                    - tasks
        '''
        return {
            'tasks': tasks.get_tasks(),
        }

    def post(self):
        '''
        Runs a module as a background task
        ---
        parameters:
          - name: body
            in: body
            description: Object containing the path of the module to run
            schema:
                properties:
                    path:
                        type: string
                required:
                - path
        responses:
            201:
                description: Object containing the ID of the created task
                schema:
                    properties:
                        task:
                            type: string
                    required:
                    - task
        '''
        path = request.json.get('path')
        if not path or path not in recon._loaded_modules:
            abort(404)
        job = current_app.task_queue.enqueue('recon.core.tasks.run_module', current_app.config['WORKSPACE'], path)
        tid = job.get_id()
        status = job.get_status()
        tasks.add_task(tid, status)
        return {
            'task': tid,
        }, 201

api.add_resource(TaskList, '/tasks/')


class TaskInst(Resource):

    def get(self, tid):
        '''
        Gets the specified task
        ---
        parameters:
          - name: tid
            in: path
            description: ID of the target task
            required: true
            type: string
          - name: live
            in: query
            description: If set, queries the Redis queue instead of the database
            required: false
            type: string
        responses:
            200:
                description: Object containing the specified task
                schema:
                    $ref: '#/definitions/Task'
        '''
        if tid not in tasks.get_ids():
            abort(404)
        # process requests for the rq version of the task
        if request.args.get('live'):
            # baseline task object
            task = {
                'id': tid,
                'status': 'unknown',
                'result': None,
            }
            job = current_app.task_queue.fetch_job(tid)
            if job:
                task['status'] = job.get_status()
                task['result'] = job.result
        else:
            task = tasks.get_task(tid)
        return task

api.add_resource(TaskInst, '/tasks/<string:tid>')


class ModuleList(Resource):

    def get(self):
        '''
        Gets all module names from the framework
        ---
        responses:
            200:
                description: List of module names
                schema:
                    properties:
                        modules:
                            type: array
                            items:
                                type: string
                    required:
                    - modules
        '''
        return {
            'modules': sorted(list(recon._loaded_modules.keys())),
        }

api.add_resource(ModuleList, '/modules/')


class ModuleInst(Resource):

    def get(self, module):
        '''
        Gets information about the specified module
        ---
        parameters:
          - name: module
            in: path
            description: Path of the target module
            required: true
            type: string
        responses:
            200:
                description: Object containing the specified module's information
                schema:
                    $ref: '#/definitions/Module'
        '''
        module = recon._loaded_modules.get(module)
        if module is None:
            abort(404)
        meta = {k: v for k, v in module.meta.items()}
        # provide options with more context
        options = module.options.serialize()
        if options:
            meta['options'] = options
        return meta

    def patch(self, module):
        '''
        Updates the specified module
        Options are the only modifiable property of a module object.
        ---
        parameters:
          - name: module
            in: path
            description: Name of the target module
            required: true
            type: string
          - name: body
            in: body
            description: Object containing the options to update
            schema:
                properties:
                    options:
                        type: array
                        items:
                            type: object
                            properties:
                                name:
                                    type: string
                                value:
                                    type: string
                            required:
                            - name
                            - value
                required:
                - options
        responses:
            200:
                description: Object containing the modified module's information
                schema:
                    $ref: '#/definitions/Module'
        '''
        module = recon._loaded_modules.get(module)
        if module is None:
            abort(404)
        options = request.json.get('options')
        # process options
        if options:
            for option in options:
                name = option.get('name')
                value = option.get('value')
                if name and value and name in module.options:
                    module.options[name] = value
                    module._save_config(name)
        return self.get(module._modulename)

api.add_resource(ModuleInst, '/modules/<path:module>')


class WorkspaceList(Resource):

    def get(self):
        '''
        Gets all workspace names from the framework
        ---
        responses:
            200:
                description: List of workspace names
                schema:
                    properties:
                        workspaces:
                            type: array
                            items:
                                type: string
                    required:
                    - workspaces
        '''
        return {
            'workspaces': sorted(recon._get_workspaces()),
        }

api.add_resource(WorkspaceList, '/workspaces/')


class WorkspaceInst(Resource):

    def get(self, workspace):
        '''
        Gets information about the specified workspace
        Only returns options for the active workspace.
        ---
        parameters:
          - name: workspace
            in: path
            description: Name of the target workspace
            required: true
            type: string
        responses:
            200:
                description: Object containing the specified workspace's information
                schema:
                    $ref: '#/definitions/Workspace'
        '''
        if workspace not in recon._get_workspaces():
            abort(404)
        status = 'inactive'
        options = []
        if workspace == current_app.config['WORKSPACE']:
            status = 'active'
            options = recon.options.serialize()
        return {
            'name': workspace,
            'status': status,
            'options': options,
        }

    def patch(self, workspace):
        '''
        Updates the specified workspace
        Activating a workspace deactivates the currently activated workspace, and only the active workspace's options can be modified.
        ---
        parameters:
          - name: workspace
            in: path
            description: Name of the target workspace
            required: true
            type: string
          - name: body
            in: body
            description: Object containing the properties to update
            schema:
                properties:
                    status:
                        type: string
                    options:
                        type: array
                        items:
                            type: object
                            properties:
                                name:
                                    type: string
                                value:
                                    type: string
                            required:
                            - name
                            - value
        responses:
            200:
                description: Object containing the modified workspace's information
                schema:
                    $ref: '#/definitions/Workspace'
        '''
        if workspace not in recon._get_workspaces():
            abort(404)
        status = request.json.get('status')
        options = request.json.get('options')
        # process status
        if status:
            # ignore everything but a request to activate
            if status == 'active':
                # only continue if the workspace is not already active
                if current_app.config['WORKSPACE'] != workspace:
                    # re-initialize the workspace and tasks object
                    recon._init_workspace(workspace)
                    tasks.__init__(recon)
                    # add the workspace name the to global object
                    current_app.config['WORKSPACE'] = workspace
                    print((f" * Workspace initialized: {workspace}"))
        # process options
        if options:
            # only continue if the workspace is active
            if current_app.config['WORKSPACE'] == workspace:
                for option in options:
                    name = option.get('name')
                    value = option.get('value')
                    if name and value and name in recon.options:
                        recon.options[name] = value
                        recon._save_config(name)
        return self.get(workspace)

api.add_resource(WorkspaceInst, '/workspaces/<string:workspace>')


class DashboardInst(Resource):

    def get(self):
        '''
        Gets summary information about the current workspace
        ---
        responses:
            200:
                description: Object containing the summary information
                schema:
                    $ref: '#/definitions/Dashboard'
        '''
        # build the activity object
        dashboard = recon.query('SELECT * FROM dashboard', include_header=True)
        columns = dashboard.pop(0)
        activity = columnize(columns, dashboard)
        # build the records object
        records = []
        tables = recon.get_tables()
        for table in tables:
            count = recon.query(f"SELECT COUNT(*) AS 'COUNT' FROM {table}")
            records.append({'name': table, 'count':count[0][0]})
        # sort both lists in descending order
        records.sort(key=lambda r: r['count'], reverse=True)
        activity.sort(key=lambda m: m['runs'], reverse=True)
        return {
            'workspace': current_app.config['WORKSPACE'],
            'records': records,
            'activity': activity,
        }

api.add_resource(DashboardInst, '/dashboard')


class ReportList(Resource):

    def get(self):
        '''
        Gets all report types from the framework
        ---
        responses:
            200:
                description: List of report types
                schema:
                    properties:
                        reports:
                            type: array
                            items:
                                type: string
                    required:
                    - reports
        '''
        return {
            'reports': sorted(list(REPORTS.keys())),
        }

api.add_resource(ReportList, '/reports/')


class ReportInst(Resource):

    def get(self, report):
        '''
        Runs the specified report for the current workspace
        ---
        parameters:
          - name: report
            in: path
            description: Name of the report type
            required: true
            type: string
        '''
        if report not in REPORTS:
            abort(404)
        return REPORTS[report]()

api.add_resource(ReportInst, '/reports/<string:report>')


class TableList(Resource):

    def get(self):
        '''
        Gets all table names for the current workspace
        ---
        responses:
            200:
                description: Object containing the list of tables names
                schema:
                    properties:
                        workspace:
                            type: string
                        tables:
                            type: array
                            items:
                                type: string
                    required:
                    - workspace
                    - tables
        '''
        return {
            'workspace': current_app.config['WORKSPACE'],
            'tables': sorted(recon.get_tables()),
        }

api.add_resource(TableList, '/tables/')


class TableInst(Resource):

    def get(self, table):
        '''
        Dumps the contents of the specified table
        ---
        parameters:
          - name: table
            in: path
            description: Name of the target table
            required: true
            type: string
          - name: format
            in: query
            description: Export type
            required: false
            type: string
        responses:
            200:
                description: Object containing the specified table's contents
                schema:
                    $ref: '#/definitions/Table'
        '''
        if table not in recon.get_tables():
            abort(404)
        # filter rows for columns if needed
        columns = request.values.get('columns')
        if columns:
            rows = recon.query(f"SELECT {columns} FROM {table}", include_header=True)
        else:
            rows = recon.query(f"SELECT * FROM {table}", include_header=True)
        columns = rows.pop(0)
        rows = columnize(columns, rows)
        # dynamically determine and call export function
        _format = request.args.get('format')
        if _format and _format in EXPORTS:
            # any required serialization is handled at the exporter level
            return EXPORTS[_format](rows=rows)
        return {
            'workspace': current_app.config['WORKSPACE'],
            'table': table,
            'columns': columns,
            'rows': rows,
        }

api.add_resource(TableInst, '/tables/<string:table>')


class ExportList(Resource):

    def get(self):
        '''
        Gets all export types from the framework
        ---
        responses:
            200:
                description: List of export types
                schema:
                    properties:
                        exports:
                            type: array
                            items:
                                type: string
                    required:
                    - exports
        '''
        return {
            'exports': sorted(list(EXPORTS.keys())),
        }

api.add_resource(ExportList, '/exports')



import hashlib
import hmac
import sys
import time


class PwnedlistMixin(object):

    def build_pwnedlist_payload(self, payload, method, key, secret):
        timestamp = int(time.time())
        payload['ts'] = timestamp
        payload['key'] = key
        msg = f"{key}{timestamp}{method}{secret}"
        encoding = sys.getdefaultencoding()
        hm = hmac.new(bytes(secret, encoding), bytes(msg, encoding), hashlib.sha1)
        payload['hmac'] = hm.hexdigest()
        return payload

    def get_pwnedlist_leak(self, leak_id):
        # check if the leak has already been retrieved
        leak = self.query('SELECT * FROM leaks WHERE leak_id=?', (leak_id,))
        if leak:
            leak = dict(zip([x[0] for x in self.get_columns('leaks')], leak[0]))
            del leak['module']
            return leak
        # set up the API call
        key = self.get_key('pwnedlist_api')
        secret = self.get_key('pwnedlist_secret')
        url = 'https://api.pwnedlist.com/api/1/leaks/info'
        base_payload = {'leakId': leak_id}
        payload = self.build_pwnedlist_payload(base_payload, 'leaks.info', key, secret)
        # make the request
        resp = self.request('GET', url, params=payload)
        if resp.status_code != 200:
            self.error(f"Error retrieving leak data.{os.linesep}{resp.text}")
            return
        leak = resp.json()['leaks'][0]
        # normalize the leak for storage
        normalized_leak = {}
        for item in leak:
            value = leak[item]
            if type(value) == list:
                value = ', '.join(value)
            normalized_leak[item] = value
        return normalized_leak



from requests.exceptions import Timeout
import html
import http.cookiejar
import io
import os
import re
import socket
import sqlite3
import sys
import textwrap
import yaml
# framework libs
from recon.core import framework
from recon.utils import validators

#=================================================
# MODULE CLASS
#=================================================

class BaseModule(framework.Framework):

    def __init__(self, params):
        framework.Framework.__init__(self, params)
        self.options = framework.Options()
        # update the meta dictionary by merging the class variable with any frontmatter
        self.meta = self._merge_dicts(self.meta, self._parse_frontmatter())
        # register a data source option if a default query is specified in the module
        if self.meta.get('query'):
            self._default_source = self.meta.get('query')
            self.register_option('source', 'default', True, 'source of input (see \'info\' for details)')
        # register all other specified options
        if self.meta.get('options'):
            for option in self.meta.get('options'):
                self.register_option(*option)
        # register any required keys
        if self.meta.get('required_keys'):
            self.keys = {}
            for key in self.meta.get('required_keys'):
                # add key to the database
                self._query_keys('INSERT OR IGNORE INTO keys (name) VALUES (?)', (key,))
                # migrate the old key if needed
                self._migrate_key(key)
                # add key to local keys dictionary
                # could fail to load on exception here to prevent loading modules
                # without required keys, but would need to do it in a separate loop
                # so that all keys get added to the database first. for now, the
                # framework will warn users of the missing key, but allow the module
                # to load.
                self.keys[key] = self.get_key(key)
                if not self.keys.get(key):
                    self.error(f"'{key}' key not set. {self._modulename.split('/')[-1]} module will likely fail at runtime. See 'keys add'.")
        self._reload = 0

    #==================================================
    # SUPPORT METHODS
    #==================================================

    def _merge_dicts(self, x, y):
        # start with x's keys and values
        z = x.copy()
        # modify z with y's keys and values
        z.update(y)
        return z

    def _parse_frontmatter(self):
        rel_path = '.'.join([self._modulename, 'py'])
        abs_path = os.path.join(self.mod_path, rel_path)
        with open(abs_path) as fp:
            state = False
            yaml_src = ''
            for line in fp:
                 if line == '---\n':
                      state = not state
                      continue
                 if state:
                      yaml_src += line
        return yaml.safe_load(yaml_src) or {}

    def _migrate_key(self, key):
        '''migrate key from old .dat file'''
        key_path = os.path.join(self.home_path, 'keys.dat')
        if os.path.exists(key_path):
            try:
                key_data = json.loads(open(key_path, 'rb').read())
                if key_data.get(key):
                    self.add_key(key, key_data.get(key))
            except:
                self.error(f"Corrupt key file. Manual migration of '{key}' required.")

    def ascii_sanitize(self, s):
        return ''.join([char for char in s if ord(char) in [10,13] + range(32, 126)])

    def html_unescape(self, s):
        '''Unescapes HTML markup and returns an unescaped string.'''
        return html.unescape(s)

    def html_escape(self, s):
        escapes = {
            '&': '&amp;',
            '"': '&quot;',
            "'": '&apos;',
            '>': '&gt;',
            '<': '&lt;',
            }
        return ''.join(escapes.get(c,c) for c in s)

    def cidr_to_list(self, string):
        import ipaddress
        return [str(ip) for ip in ipaddress.ip_network(string)]

    def hosts_to_domains(self, hosts, exclusions=[]):
        domains = []
        for host in hosts:
            elements = host.split('.')
            # recursively walk through the elements
            # extracting all possible (sub)domains
            while len(elements) >= 2:
                # account for domains stored as hosts
                if len(elements) == 2:
                    domain = '.'.join(elements)
                else:
                    # drop the host element
                    domain = '.'.join(elements[1:])
                if domain not in domains + exclusions:
                    domains.append(domain)
                del elements[0]
        return domains

    def _validate_input(self):
        validator_type = self.meta.get('validator')
        if not validator_type:
            # passthru, no validator required
            self.debug('No validator required.')
            return
        validator = None
        validator_name = validator_type.capitalize() + 'Validator'
        for obj in [self, validators]:
            if hasattr(obj, validator_name):
                validator = getattr(validators, validator_name)()
        if not validator:
            # passthru, no validator defined
            self.debug('No validator defined.')
            return
        inputs = self._get_source(self.options['source'], self._default_source)
        for _input in inputs:
            validator.validate(_input)
            self.debug('All inputs validated.')

    #==================================================
    # OPTIONS METHODS
    #==================================================

    def _get_source(self, params, query=None):
        prefix = params.split()[0].lower()
        if prefix in ['query', 'default']:
            query = ' '.join(params.split()[1:]) if prefix == 'query' else query
            try:
                results = self.query(query)
            except sqlite3.OperationalError as e:
                raise framework.FrameworkException(f"Invalid source query. {type(e).__name__} {e}")
            if not results:
                sources = []
            elif len(results[0]) > 1:
                sources = [x[:len(x)] for x in results]
                #raise framework.FrameworkException('Too many columns of data as source input.')
            else:
                sources = [x[0] for x in results]
        elif os.path.exists(params):
            sources = open(params).read().split()
        else:
            sources = [params]
        if not sources:
            raise framework.FrameworkException('Source contains no input.')
        return sources

    #==================================================
    # REQUEST METHODS
    #==================================================

    def make_cookie(self, name, value, domain, path='/'):
        return http.cookiejar.Cookie(
            version=0, 
            name=name, 
            value=value,
            port=None, 
            port_specified=False,
            domain=domain, 
            domain_specified=True, 
            domain_initial_dot=False,
            path=path, 
            path_specified=True,
            secure=False,
            expires=None,
            discard=False,
            comment=None,
            comment_url=None,
            rest=None
        )

    #==================================================
    # COMMAND METHODS
    #==================================================

    def do_goptions(self, params):
        '''Manages the global context options'''
        if not params:
            self.help_goptions()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('goptions'):
            return getattr(self, '_do_goptions_'+arg)(params)
        else:
            self.help_goptions()

    def _do_goptions_list(self, params):
        '''Shows the global context options'''
        self._list_options(self._global_options)

    def _do_goptions_set(self, params):
        '''Sets a global context option'''
        option, value = self._parse_params(params)
        if not (option and value):
            self._help_goptions_set()
            return
        name = option.upper()
        if name in self._global_options:
            self._global_options[name] = value
            print(f"{name} => {value}")
            self._save_config(name, 'base', self._global_options)
        else:
            self.error('Invalid option name.')

    def _do_goptions_unset(self, params):
        '''Unsets a global context option'''
        option, value = self._parse_params(params)
        if not option:
            self._help_goptions_unset()
            return
        name = option.upper()
        if name in self._global_options:
            self._do_goptions_set(' '.join([name, 'None']))
        else:
            self.error('Invalid option name.')

    def _do_modules_load(self, params):
        '''Loads a module'''
        if not params:
            self._help_modules_load()
            return
        # finds any modules that contain params
        modules = self._match_modules(params)
        # notify the user if none or multiple modules are found
        if len(modules) != 1:
            if not modules:
                self.error('Invalid module name.')
            else:
                self.output(f"Multiple modules match '{params}'.")
                self._list_modules(modules)
            return
        # compensation for stdin being used for scripting and loading
        if framework.Framework._script:
            end_string = sys.stdin.read()
        else:
            end_string = 'EOF'
            framework.Framework._load = 1
        sys.stdin = io.StringIO(f"modules load {modules[0]}{os.linesep}{end_string}")
        return True

    def do_reload(self, params):
        '''Reloads the loaded module'''
        self._reload = 1
        return True

    def do_info(self, params):
        '''Shows details about the loaded module'''
        print('')
        # meta info
        for item in ['name', 'author', 'version']:
            print(f"{item.title().rjust(10)}: {self.meta[item]}")
        # required keys
        if self.meta.get('required_keys'):
            print(f"{'keys'.title().rjust(10)}: {', '.join(self.meta.get('required_keys'))}")
        print('')
        # description
        print('Description:')
        print(f"{self.spacer}{textwrap.fill(self.meta['description'], 100, subsequent_indent=self.spacer)}")
        print('')
        # options
        print('Options:', end='')
        self._list_options()
        # sources
        if hasattr(self, '_default_source'):
            print('Source Options:')
            print(f"{self.spacer}{'default'.ljust(15)}{self._default_source}")
            print(f"{self.spacer}{'<string>'.ljust(15)}string representing a single input")
            print(f"{self.spacer}{'<path>'.ljust(15)}path to a file containing a list of inputs")
            print(f"{self.spacer}{'query <sql>'.ljust(15)}database query returning one column of inputs")
            print('')
        # comments
        if self.meta.get('comments'):
            print('Comments:')
            for comment in self.meta['comments']:
                prefix = '* '
                if comment.startswith('\t'):
                    prefix = self.spacer+'- '
                    comment = comment[1:]
                print(f"{self.spacer}{textwrap.fill(prefix+comment, 100, subsequent_indent=self.spacer)}")
            print('')

    def do_input(self, params):
        '''Shows inputs based on the source option'''
        if hasattr(self, '_default_source'):
            try:
                self._validate_options()
                inputs = self._get_source(self.options['source'], self._default_source)
                self.table([[x] for x in inputs], header=['Module Inputs'])
            except Exception as e:
                self.output(e.__str__())
        else:
            self.output('Source option not available for this module.')

    def run(self):
        self._validate_options()
        self._validate_input()
        self._summary_counts = {}
        pre = self.module_pre()
        params = [pre] if pre is not None else []
        # provide input if a default query is specified in the module
        if hasattr(self, '_default_source'):
            objs = self._get_source(self.options['source'], self._default_source)
            params.insert(0, objs)
        # update the dashboard before running the module
        # data is added at runtime, so even if an error occurs, any new items
        # must be accounted for by a module execution attempt
        self.query(f"INSERT OR REPLACE INTO dashboard (module, runs) VALUES ('{self._modulename}', COALESCE((SELECT runs FROM dashboard WHERE module='{self._modulename}')+1, 1))")
        self.module_run(*params)
        self.module_post()

    def do_run(self, params):
        '''Runs the loaded module'''
        try:
            self.run()
        except KeyboardInterrupt:
            print('')
        except (Timeout, socket.timeout):
            self.print_exception()
            self.error('A request took too long to complete. If the issue persists, increase the global TIMEOUT option.')
        except (framework.FrameworkException, validators.ValidationException):
            self.print_exception()
        except Exception:
            self.print_exception()
            self.error('Something broken? See https://github.com/lanmaster53/recon-ng/wiki/Troubleshooting#issue-reporting.')
        finally:
            # print module summary
            if self._summary_counts:
                self.heading('Summary', level=0)
                for table in self._summary_counts:
                    new = self._summary_counts[table]['new']
                    cnt = self._summary_counts[table]['count']
                    if new > 0:
                        method = getattr(self, 'alert')
                    else:
                        method = getattr(self, 'output')
                    method(f"{cnt} total ({new} new) {table} found.")

    #==================================================
    # HELP METHODS
    #==================================================

    def help_goptions(self):
        print(getattr(self, 'do_goptions').__doc__)
        print(f"{os.linesep}Usage: goptions <{'|'.join(self._parse_subcommands('goptions'))}> [...]{os.linesep}")

    def _help_goptions_set(self):
        print(getattr(self, '_do_goptions_set').__doc__)
        print(f"{os.linesep}Usage: goptions set <option> <value>{os.linesep}")

    def _help_goptions_unset(self):
        print(getattr(self, '_do_goptions_unset').__doc__)
        print(f"{os.linesep}Usage: goptions unset <option>{os.linesep}")

    #==================================================
    # COMPLETE METHODS
    #==================================================

    def complete_goptions(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('goptions')
        if arg in subs:
            return getattr(self, '_complete_goptions_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_goptions_list(self, text, *ignored):
        return []

    def _complete_goptions_set(self, text, *ignored):
        return [x for x in self._global_options if x.startswith(text.upper())]
    _complete_goptions_unset = _complete_goptions_set

    def complete_reload(self, text, *ignored):
        return []
    complete_info = complete_input = complete_run = complete_reload

    #==================================================
    # HOOK METHODS
    #==================================================

    def module_pre(self):
        pass

    def module_run(self):
        pass

    def module_post(self):
        pass



from contextlib import closing
import cmd
import codecs
import inspect
import json
import os
import random
import re
import requests
import socket
import sqlite3
import string
import subprocess
import sys
import traceback

#=================================================
# SUPPORT CLASSES
#=================================================

class FrameworkException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)

class Colors(object):
    N = '\033[m' # native
    R = '\033[31m' # red
    G = '\033[32m' # green
    O = '\033[33m' # orange
    B = '\033[34m' # blue

class Options(dict):

    def __init__(self, *args, **kwargs):
        self.required = {}
        self.description = {}
        super(Options, self).__init__(*args, **kwargs)

    def __getitem__(self, name):
        name = self.__keytransform__(name)
        return super(Options, self).__getitem__(name)

    def __setitem__(self, name, value):
        name = self.__keytransform__(name)
        value = self._autoconvert(value)
        super(Options, self).__setitem__(name, value)

    def __delitem__(self, name):
        name = self.__keytransform__(name)
        super(Options, self).__delitem__(name)
        if name in self.required:
            del self.required[name]
        if name in self.description:
            del self.description[name]

    def __keytransform__(self, key):
        return key.upper()

    def _boolify(self, value):
        # designed to throw an exception if value is not a string representation of a boolean
        return {'true':True, 'false':False}[value.lower()]

    def _autoconvert(self, value):
        if value in (None, True, False):
            return value
        elif (isinstance(value, str)) and value.lower() in ('none', "''", '""'):
            return None
        orig = value
        for fn in (self._boolify, int, float):
            try:
                value = fn(value)
                break
            except ValueError: pass
            except KeyError: pass
            except AttributeError: pass
        if type(value) is int and '.' in str(orig):
            return float(orig)
        return value

    def init_option(self, name, value=None, required=False, description=''):
        name = self.__keytransform__(name)
        self[name] = value
        self.required[name] = required
        self.description[name] = description

    def serialize(self):
        options = []
        for key in self:
            option = {}
            option['name'] = key
            option['value'] = self[key]
            option['required'] = self.required[key]
            option['description'] = self.description[key]
            options.append(option)
        return options

#=================================================
# FRAMEWORK CLASS
#=================================================

class Framework(cmd.Cmd):
    prompt = '>>>'
    # mode flags
    _script = 0
    _load = 0
    _mode = 0
    # framework variables
    _global_options = Options()
    _loaded_modules = {}
    app_path = ''
    data_path = ''
    core_path = ''
    home_path = ''
    mod_path = ''
    spaces_path = ''
    workspace = ''
    _record = None
    _spool = None
    _summary_counts = {}

    def __init__(self, params):
        cmd.Cmd.__init__(self)
        self._modulename = params
        self.ruler = '-'
        self.spacer = '  '
        self.time_format = '%Y-%m-%d %H:%M:%S'
        self.nohelp = f"{Colors.R}[!] No help on %s{Colors.N}"
        self.do_help.__func__.__doc__ = '''Displays this menu'''
        self.doc_header = 'Commands (type [help|?] <topic>):'
        self._exit = 0

    #==================================================
    # CMD OVERRIDE METHODS
    #==================================================

    def default(self, line):
        self.error(f"Invalid command: {line}")

    def emptyline(self):
        # disables running of last command when no command is given
        # return flag to tell interpreter to continue
        return 0

    def precmd(self, line):
        if Framework._load:
            print('\r', end='')
        if Framework._script:
            print(f"{line}")
        if Framework._record:
            recorder = codecs.open(Framework._record, 'ab', encoding='utf-8')
            recorder.write(f"{line}{os.linesep}")
            recorder.flush()
            recorder.close()
        if Framework._spool:
            Framework._spool.write(f"{self.prompt}{line}{os.linesep}")
            Framework._spool.flush()
        return line

    def onecmd(self, line):
        cmd, arg, line = self.parseline(line)
        if not line:
            return self.emptyline()
        if line == 'EOF':
            # reset stdin for raw_input
            sys.stdin = sys.__stdin__
            Framework._script = 0
            Framework._load = 0
            return
        if cmd is None:
            return self.default(line)
        self.lastcmd = line
        if cmd == '':
            return self.default(line)
        else:
            try:
                func = getattr(self, 'do_' + cmd)
            except AttributeError:
                return self.default(line)
            try:
                return func(arg)
            except Exception:
                self.print_exception()

    # make help menu more attractive
    def print_topics(self, header, cmds, cmdlen, maxcol):
        if cmds:
            self.stdout.write(f"{header}{os.linesep}")
            if self.ruler:
                self.stdout.write(f"{self.ruler * len(header)}{os.linesep}")
            for cmd in cmds:
                self.stdout.write(f"{cmd.ljust(15)} {getattr(self, 'do_' + cmd).__doc__}{os.linesep}")
            self.stdout.write(os.linesep)

    #==================================================
    # SUPPORT METHODS
    #==================================================

    def to_unicode_str(self, obj, encoding='utf-8'):
        # converts non-stringish types to unicode
        if type(obj) not in (str, bytes):
            obj = str(obj)
        obj = self.to_unicode(obj, encoding)
        return obj

    def to_unicode(self, obj, encoding='utf-8'):
        # converts bytes to unicode
        if isinstance(obj, bytes):
            obj = obj.decode(encoding)
        return obj

    def is_hash(self, hashstr):
        hashdict = [
            {'pattern': r'^[a-fA-F0-9]{32}$', 'type': 'MD5'},
            {'pattern': r'^[a-fA-F0-9]{16}$', 'type': 'MySQL'},
            {'pattern': r'^\*[a-fA-F0-9]{40}$', 'type': 'MySQL5'},
            {'pattern': r'^[a-fA-F0-9]{40}$', 'type': 'SHA1'},
            {'pattern': r'^[a-fA-F0-9]{56}$', 'type': 'SHA224'},
            {'pattern': r'^[a-fA-F0-9]{64}$', 'type': 'SHA256'},
            {'pattern': r'^[a-fA-F0-9]{96}$', 'type': 'SHA384'},
            {'pattern': r'^[a-fA-F0-9]{128}$', 'type': 'SHA512'},
            {'pattern': r'^\$[PH]{1}\$.{31}$', 'type': 'phpass'},
            {'pattern': r'^\$2[ya]?\$.{56}$', 'type': 'bcrypt'},
        ]
        for hashitem in hashdict:
            if re.match(hashitem['pattern'], hashstr):
                return hashitem['type']
        return False

    def get_random_str(self, length):
        return ''.join(random.choice(string.ascii_lowercase) for i in range(length))

    def _is_writeable(self, filename):
        try:
            fp = open(filename, 'a')
            fp.close()
            return True
        except IOError:
            return False

    def _parse_rowids(self, rowids):
        xploded = []
        rowids = [x.strip() for x in rowids.split(',')]
        for rowid in rowids:
            try:
                if '-' in rowid:
                    start = int(rowid.split('-')[0].strip())
                    end = int(rowid.split('-')[-1].strip())
                    xploded += range(start, end+1)
                else:
                    xploded.append(int(rowid))
            except ValueError:
                continue
        return sorted(list(set(xploded)))

    #==================================================
    # OUTPUT METHODS
    #==================================================

    def print_exception(self, line=''):
        stack_list = [x.strip() for x in traceback.format_exc().strip().splitlines()]
        exctype = stack_list[-1].split(':', 1)[0].strip()
        message = stack_list[-1].split(':', 1)[-1].strip()
        if self._global_options['verbosity'] == 0:
            return
        elif self._global_options['verbosity'] == 1:
            line = ' '.join([x for x in [message, line] if x])
            self.error(line)
        elif self._global_options['verbosity'] == 2:
            print(f"{Colors.R}{'-'*60}")
            traceback.print_exc()
            print(f"{'-'*60}{Colors.N}")

    def error(self, line):
        '''Formats and presents errors.'''
        if not re.search('[.,;!?]$', line):
            line += '.'
        line = line[:1].upper() + line[1:]
        print(f"{Colors.R}[!] {line}{Colors.N}")

    def output(self, line):
        '''Formats and presents normal output.'''
        print(f"{Colors.B}[*]{Colors.N} {line}")

    def alert(self, line):
        '''Formats and presents important output.'''
        print(f"{Colors.G}[*]{Colors.N} {line}")

    def verbose(self, line):
        '''Formats and presents output if in verbose mode.'''
        if self._global_options['verbosity'] >= 1:
            self.output(line)

    def debug(self, line):
        '''Formats and presents output if in debug mode (very verbose).'''
        if self._global_options['verbosity'] >= 2:
            self.output(line)

    def heading(self, line, level=1):
        '''Formats and presents styled header text'''
        line = line
        print('')
        if level == 0:
            print(self.ruler*len(line))
            print(line.upper())
            print(self.ruler*len(line))
        if level == 1:
            print(f"{self.spacer}{line.title()}")
            print(f"{self.spacer}{self.ruler*len(line)}")

    def table(self, data, header=[], title=''):
        '''Accepts a list of rows and outputs a table.'''
        tdata = list(data)
        if header:
            tdata.insert(0, header)
        if len(set([len(x) for x in tdata])) > 1:
            raise FrameworkException('Row lengths not consistent.')
        lens = []
        cols = len(tdata[0])
        # create a list of max widths for each column
        for i in range(0,cols):
            lens.append(len(max([self.to_unicode_str(x[i]) if x[i] != None else '' for x in tdata], key=len)))
        # calculate dynamic widths based on the title
        title_len = len(title)
        tdata_len = sum(lens) + (3*(cols-1))
        diff = title_len - tdata_len
        if diff > 0:
            diff_per = diff / cols
            lens = [x+diff_per for x in lens]
            diff_mod = diff % cols
            for x in range(0, diff_mod):
                lens[x] += 1
        # build ascii table
        if len(tdata) > 0:
            separator_str = f"{self.spacer}+-{'%s---'*(cols-1)}%s-+"
            separator_sub = tuple(['-'*x for x in lens])
            separator = separator_str % separator_sub
            data_str = f"{self.spacer}| {'%s | '*(cols-1)}%s |"
            # top of ascii table
            print('')
            print(separator)
            # ascii table data
            if title:
                print(f"{self.spacer}| {title.center(tdata_len)} |")
                print(separator)
            if header:
                rdata = tdata.pop(0)
                data_sub = tuple([rdata[i].center(lens[i]) for i in range(0,cols)])
                print(data_str % data_sub)
                print(separator)
            for rdata in tdata:
                data_sub = tuple([self.to_unicode_str(rdata[i]).ljust(lens[i]) if rdata[i] != None else ''.ljust(lens[i]) for i in range(0,cols)])
                print(data_str % data_sub)
            # bottom of ascii table
            print(separator)
            print('')

    #==================================================
    # DATABASE METHODS
    #==================================================

    def query(self, *args, **kwargs):
        path = os.path.join(self.workspace, 'data.db')
        return self._query(path, *args, **kwargs)

    def _query(self, path, query, values=(), include_header=False):
        '''Queries the database and returns the results as a list.'''
        self.debug(f"DATABASE => {path}")
        self.debug(f"QUERY => {query}")
        with sqlite3.connect(path) as conn:
            with closing(conn.cursor()) as cur:
                if values:
                    self.debug(f"VALUES => {repr(values)}")
                    cur.execute(query, values)
                else:
                    cur.execute(query)
                # a rowcount of -1 typically refers to a select statement
                if cur.rowcount == -1:
                    rows = []
                    if include_header:
                        rows.append(tuple([x[0] for x in cur.description]))
                    rows.extend(cur.fetchall())
                    results = rows
                # a rowcount of 1 == success and 0 == failure
                else:
                    conn.commit()
                    results = cur.rowcount
                return results

    def get_columns(self, table):
        return [(x[1], x[2]) for x in self.query(f"PRAGMA table_info('{table}')")]

    def get_tables(self):
        return [x[0] for x in self.query('SELECT name FROM sqlite_master WHERE type=\'table\'') if x[0] not in ['dashboard']]

    #==================================================
    # INSERT METHODS
    #==================================================

    def _display(self, data, rowcount):
        display = self.alert if rowcount else self.verbose
        for key in sorted(data.keys()):
            display(f"{key.title()}: {data[key]}")
        display(self.ruler*50)

    def insert_domains(self, domain=None, notes=None, mute=False):
        '''Adds a domain to the database and returns the affected row count.'''
        data = dict(
            domain = domain,
            notes = notes
        )
        rowcount = self.insert('domains', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_companies(self, company=None, description=None, notes=None, mute=False):
        '''Adds a company to the database and returns the affected row count.'''
        data = dict(
            company = company,
            description = description,
            notes = notes
        )
        rowcount = self.insert('companies', data.copy(), ('company',))
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_netblocks(self, netblock=None, notes=None, mute=False):
        '''Adds a netblock to the database and returns the affected row count.'''
        data = dict(
            netblock = netblock,
            notes = notes
        )
        rowcount = self.insert('netblocks', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_locations(self, latitude=None, longitude=None, street_address=None, notes=None, mute=False):
        '''Adds a location to the database and returns the affected row count.'''
        data = dict(
            latitude = latitude,
            longitude = longitude,
            street_address = street_address,
            notes = notes
        )
        rowcount = self.insert('locations', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_vulnerabilities(self, host=None, reference=None, example=None, publish_date=None, category=None, status=None, notes=None, mute=False):
        '''Adds a vulnerability to the database and returns the affected row count.'''
        data = dict(
            host = host,
            reference = reference,
            example = example,
            publish_date = publish_date.strftime(self.time_format) if publish_date else None,
            category = category,
            status = status,
            notes = notes
        )
        rowcount = self.insert('vulnerabilities', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_ports(self, ip_address=None, host=None, port=None, protocol=None, banner=None, notes=None, mute=False):
        '''Adds a port to the database and returns the affected row count.'''
        data = dict(
            ip_address = ip_address,
            port = port,
            host = host,
            protocol = protocol,
            banner = banner,
            notes = notes
        )
        rowcount = self.insert('ports', data.copy(), ('ip_address', 'port', 'host'))
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_hosts(self, host=None, ip_address=None, region=None, country=None, latitude=None, longitude=None, notes=None, mute=False):
        '''Adds a host to the database and returns the affected row count.'''
        data = dict(
            host = host,
            ip_address = ip_address,
            region = region,
            country = country,
            latitude = latitude,
            longitude = longitude,
            notes = notes
        )
        rowcount = self.insert('hosts', data.copy(), ('host', 'ip_address'))
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_contacts(self, first_name=None, middle_name=None, last_name=None, email=None, title=None, region=None, country=None, phone=None, notes=None, mute=False):
        '''Adds a contact to the database and returns the affected row count.'''
        data = dict(
            first_name = first_name,
            middle_name = middle_name,
            last_name = last_name,
            title = title,
            email = email,
            region = region,
            country = country,
            phone = phone,
            notes = notes
        )
        rowcount = self.insert('contacts', data.copy(), ('first_name', 'middle_name', 'last_name', 'title', 'email', 'phone'))
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_credentials(self, username=None, password=None, _hash=None, _type=None, leak=None, notes=None, mute=False):
        '''Adds a credential to the database and returns the affected row count.'''

        # account for hashes provided in the password field
        if password and not _hash:
            _type = self.is_hash(password)
            if _type:
                _hash = password
                password = None
        # handle hashes provided without a type
        if _hash and not _type:
            _type = self.is_hash(_hash)
        # add email usernames to contacts
        if username is not None and '@' in username:
            self.insert_contacts(first_name=None, last_name=None, title=None, email=username)

        data = dict (
            username = username,
            password = password,
            hash = _hash,
            type = _type,
            leak = leak,
            notes = notes
        )
        rowcount = self.insert('credentials', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_leaks(self, leak_id=None, description=None, source_refs=None, leak_type=None, title=None, import_date=None, leak_date=None, attackers=None, num_entries=None, score=None, num_domains_affected=None, attack_method=None, target_industries=None, password_hash=None, password_type=None, targets=None, media_refs=None, notes=None, mute=False):
        '''Adds a leak to the database and returns the affected row count.'''
        data = dict(
            leak_id = leak_id,
            description = description,
            source_refs = source_refs,
            leak_type = leak_type,
            title = title,
            import_date = import_date,
            leak_date = leak_date,
            attackers = attackers,
            num_entries = num_entries,
            score = score,
            num_domains_affected = num_domains_affected,
            attack_method = attack_method,
            target_industries = target_industries,
            password_hash = password_hash,
            password_type = password_type,
            targets = targets,
            media_refs = media_refs,
            notes = notes
        )
        rowcount = self.insert('leaks', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_pushpins(self, source=None, screen_name=None, profile_name=None, profile_url=None, media_url=None, thumb_url=None, message=None, latitude=None, longitude=None, time=None, notes=None, mute=False):
        '''Adds a pushpin to the database and returns the affected row count.'''
        data = dict(
            source = source,
            screen_name = screen_name,
            profile_name = profile_name,
            profile_url = profile_url,
            media_url = media_url,
            thumb_url = thumb_url,
            message = message,
            latitude = latitude,
            longitude = longitude,
            time = time.strftime(self.time_format),
            notes = notes
        )
        rowcount = self.insert('pushpins', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_profiles(self, username=None, resource=None, url=None, category=None, notes=None, mute=False):
        '''Adds a profile to the database and returns the affected row count.'''
        data = dict(
            username = username,
            resource = resource,
            url = url,
            category = category,
            notes = notes
        )
        rowcount = self.insert('profiles', data.copy(), ('username', 'url'))
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert_repositories(self, name=None, owner=None, description=None, resource=None, category=None, url=None, notes=None, mute=False):
        '''Adds a repository to the database and returns the affected row count.'''
        data = dict(
            name = name,
            owner = owner,
            description = description,
            resource = resource,
            category = category,
            url = url,
            notes = notes
        )
        rowcount = self.insert('repositories', data.copy(), data.keys())
        if not mute: self._display(data, rowcount)
        return rowcount

    def insert(self, table, data, unique_columns=[]):
        '''Inserts items into database and returns the affected row count.
        table - the table to insert the data into
        data - the information to insert into the database table in the form of a dictionary
               where the keys are the column names and the values are the column values
        unique_columns - a list of column names that should be used to determine if the
                         information being inserted is unique'''
        # set module to the calling module unless the do_add command was used
        data['module'] = 'user_defined' if '_do_db_insert' in [x[3] for x in inspect.stack()] else self._modulename.split('/')[-1]
        # sanitize the inputs to remove NoneTypes, blank strings, and zeros
        columns = [x for x in data.keys() if data[x]]
        # make sure that module is not seen as a unique column
        unique_columns = [x for x in unique_columns if x in columns and x != 'module']
        # exit if there is nothing left to insert
        if not columns:
            return 0
        # convert any type to unicode (str) for external processing
        for column in columns:
            data[column] = self.to_unicode_str(data[column])

        # build the insert query
        columns_str = '`, `'.join(columns)
        placeholder_str = ', '.join('?'*len(columns))
        unique_columns_str = ' and '.join([f"`{column}`=?" for column in unique_columns])
        if not unique_columns:
            query = f"INSERT INTO `{table}` (`{columns_str}`) VALUES ({placeholder_str})"
        else:
            query = f"INSERT INTO `{table}` (`{columns_str}`) SELECT {placeholder_str} WHERE NOT EXISTS(SELECT * FROM `{table}` WHERE {unique_columns_str})"
        values = tuple([data[column] for column in columns] + [data[column] for column in unique_columns])

        # query the database
        rowcount = self.query(query, values)

        # increment summary tracker
        if table not in self._summary_counts:
            self._summary_counts[table] = {'count': 0, 'new': 0}
        self._summary_counts[table]['new'] += rowcount
        self._summary_counts[table]['count'] += 1

        return rowcount

    #==================================================
    # OPTIONS METHODS
    #==================================================

    def register_option(self, name, value, required, description):
        self.options.init_option(name=name, value=value, required=required, description=description)
        # needs to be optimized rather than ran on every register
        self._load_config()

    def _validate_options(self):
        for option in self.options:
            # if value type is bool or int, then we know the options is set
            if not type(self.options[option]) in [bool, int]:
                if self.options.required[option] is True and not self.options[option]:
                    raise FrameworkException(f"Value required for the '{option}' option.")
        return

    def _list_options(self, options=None):
        '''Lists options'''
        if options is None:
            options = self.options
        if options:
            pattern = f"{self.spacer}%s  %s  %s  %s"
            key_len = len(max(options, key=len))
            if key_len < 4: key_len = 4
            val_len = len(max([self.to_unicode_str(options[x]) for x in options], key=len))
            if val_len < 13: val_len = 13
            print('')
            print(pattern % ('Name'.ljust(key_len), 'Current Value'.ljust(val_len), 'Required', 'Description'))
            print(pattern % (self.ruler*key_len, (self.ruler*13).ljust(val_len), self.ruler*8, self.ruler*11))
            for key in sorted(options):
                value = options[key] if options[key] != None else ''
                reqd = 'no' if options.required[key] is False else 'yes'
                desc = options.description[key]
                print(pattern % (key.ljust(key_len), self.to_unicode_str(value).ljust(val_len), self.to_unicode_str(reqd).ljust(8), desc))
            print('')
        else:
            print('')
            print(f"{self.spacer}No options available for this module.")
            print('')

    def _load_config(self):
        config_path = os.path.join(self.workspace, 'config.dat')
        # don't bother loading if a config file doesn't exist
        if os.path.exists(config_path):
            # retrieve saved config data
            with open(config_path) as config_file:
                try:
                    config_data = json.loads(config_file.read())
                except ValueError:
                    # file is corrupt, nothing to load, exit gracefully
                    pass
                else:
                    # set option values
                    for key in self.options:
                        try:
                            self.options[key] = config_data[self._modulename][key]
                        except KeyError:
                            # invalid key, contnue to load valid keys
                            continue

    def _save_config(self, name, module=None, options=None):
        config_path = os.path.join(self.workspace, 'config.dat')
        # create a config file if one doesn't exist
        open(config_path, 'a').close()
        # retrieve saved config data
        with open(config_path) as config_file:
            try:
                config_data = json.loads(config_file.read())
            except ValueError:
                # file is empty or corrupt, nothing to load
                config_data = {}
        # override implicit defaults if specified
        module = module or self._modulename
        options = options or self.options
        # create a container for the current module
        if module not in config_data:
            config_data[module] = {}
        # set the new option value in the config
        config_data[module][name] = options[name]
        # remove the option if it has been unset
        if config_data[module][name] is None:
            del config_data[module][name]
        # remove the module container if it is empty
        if not config_data[module]:
            del config_data[module]
        # write the new config data to the config file
        with open(config_path, 'w') as config_file:
            json.dump(config_data, config_file, indent=4)

    #==================================================
    # API KEY METHODS
    #==================================================

    def get_key(self, name):
        rows = self._query_keys('SELECT value FROM keys WHERE name=? AND value NOT NULL', (name,))
        if not rows:
            return None
        return rows[0][0]

    def add_key(self, name, value):
        result = self._query_keys('UPDATE keys SET value=? WHERE name=?', (value, name))
        if not result:
            return self._query_keys('INSERT INTO keys VALUES (?, ?)', (name, value))
        return result

    def remove_key(self, name):
        #return self._query_keys('UPDATE keys SET value=NULL WHERE name=?', (name,))
        return self._query_keys('DELETE FROM keys WHERE name=?', (name,))

    def _query_keys(self, query, values=()):
        path = os.path.join(self.home_path, 'keys.db')
        result = self._query(path, query, values)
        # filter out tokens when not called from the get_key method
        if type(result) is list and 'get_key' not in [x[3] for x in inspect.stack()]:
            result = [x for x in result if not x[0].endswith('_token')]
        return result

    def _list_keys(self):
        keys = self._query_keys('SELECT * FROM keys')
        tdata = []
        for key in sorted(keys):
            tdata.append(key)
        if tdata:
            self.table(tdata, header=['Name', 'Value'])

    def _get_key_names(self):
        return [x[0] for x in self._query_keys('SELECT name FROM keys')]

    #==================================================
    # REQUEST METHODS
    #==================================================

    def _print_prepared_request(self, prepared):
        self.debug(f"{'='*25} REQUEST {'='*25}")
        print(f"url:    {prepared.url}")
        print(f"method: {prepared.method} {prepared.path_url}")
        for k, v in prepared.headers.items():
            print(f"header: {k}: {v}")
        if prepared.body:
            print(f"body: {prepared.body}")

    def _print_response(self, resp):
        self.debug(f"{'='*25} RESPONSE {'='*25}")
        print(f"status: {resp.status_code} {resp.reason}")
        for k, v in resp.headers.items():
            print(f"header: {k}: {v}")
        if resp.content:
            print(f"body:   {resp.content}")

    def request(self, method, url, **kwargs):
        # process socket timeout
        kwargs['timeout'] = kwargs.get('timeout') or self._global_options['timeout']
        # process headers
        headers = kwargs.get('headers') or {}
        # set the User-Agent header
        if 'user-agent' not in [h.lower() for h in headers]:
            headers['user-agent'] = self._global_options['user-agent']
        # normalize capitalization of the User-Agent header
        headers = {k.title(): v for k, v in headers.items()}
        kwargs['headers'] = headers
        # process proxy
        proxy = self._global_options['proxy']
        if proxy:
            proxies = {
                'http': f"http://{proxy}",
                'https': f"http://{proxy}",
            }
            kwargs['proxies'] = proxies
        # disable TLS validation and warning
        kwargs['verify'] = False
        requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)
        # send the request
        resp = getattr(requests, method.lower())(url, **kwargs)
        if self._global_options['verbosity'] < 2:
            return resp
        # display request data
        self._print_prepared_request(resp.request)
        # display response data
        self._print_response(resp)
        return resp

    #==================================================
    # MODULES METHODS
    #==================================================

    def _match_modules(self, params):
        # return an exact match
        if params in Framework._loaded_modules:
            return [params]
        # use the provided name as a keyword search and return the results
        return [x for x in Framework._loaded_modules if params in x]

    def _list_modules(self, modules):
        if modules:
            key_len = len(max(modules, key=len)) + len(self.spacer)
            last_category = ''
            for module in sorted(modules):
                category = module.split('/')[0]
                if category != last_category:
                    # print header
                    last_category = category
                    self.heading(last_category)
                # print module
                print(f"{self.spacer*2}{module}")
        else:
            print('')
            self.alert('No modules enabled/installed.')
        print('')

    #==================================================
    # SHOW METHODS
    #==================================================

    def _get_show_names(self):
        # Any method beginning with "show_" will be parsed
        # and added as a subcommand for the show command.
        prefix = 'show_'
        return [x[len(prefix):] for x in self.get_names() if x.startswith(prefix)]

    #==================================================
    # COMMAND METHODS
    #==================================================

    def _parse_subcommands(self, command):
        subcommands = []
        for method in dir(self):
            if f"_do_{command}_" in method:
                subcommands.append(method.split('_')[-1])
        return subcommands

    def _parse_params(self, params):
        params = params.split()
        arg = ''
        if params:
            arg = params.pop(0)
        params = ' '.join(params)
        return arg, params

    def do_exit(self, params):
        '''Exits the framework'''
        self._exit = 1
        return True

    def do_back(self, params):
        '''Exits the current context'''
        return True

    def do_options(self, params):
        '''Manages the current context options'''
        if not params:
            self.help_options()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('options'):
            return getattr(self, '_do_options_'+arg)(params)
        else:
            self.help_options()

    def _do_options_list(self, params):
        '''Shows the current context options'''
        self._list_options()

    def _do_options_set(self, params):
        '''Sets a current context option'''
        option, value = self._parse_params(params)
        if not (option and value):
            self._help_options_set()
            return
        name = option.upper()
        if name in self.options:
            self.options[name] = value
            print(f"{name} => {value}")
            self._save_config(name)
        else:
            self.error('Invalid option name.')

    def _do_options_unset(self, params):
        '''Unsets a current context option'''
        option, value = self._parse_params(params)
        if not option:
            self._help_options_unset()
            return
        name = option.upper()
        if name in self.options:
            self._do_options_set(' '.join([name, 'None']))
        else:
            self.error('Invalid option name.')

    def do_keys(self, params):
        '''Manages third party resource credentials'''
        if not params:
            self.help_keys()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('keys'):
            return getattr(self, '_do_keys_'+arg)(params)
        else:
            self.help_keys()

    def _do_keys_list(self, params):
        '''Lists third party resource credentials'''
        self._list_keys()

    def _do_keys_add(self, params):
        '''Adds/Updates a third party resource credential'''
        key, value = self._parse_params(params)
        if not (key and value):
            self._help_keys_add()
            return
        if self.add_key(key, value):
            self.output(f"Key '{key}' added.")

    def _do_keys_remove(self, params):
        '''Removes a third party resource credential'''
        key, value = self._parse_params(params)
        if not key:
            self._help_keys_remove()
            return
        if self.get_key(key):
            if self.remove_key(key):
                self.output(f"Key '{key}' removed.")
        else:
            self.error('Invalid key name.')

    def do_modules(self, params):
        '''Interfaces with installed modules'''
        if not params:
            self.help_modules()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('modules'):
            return getattr(self, '_do_modules_'+arg)(params)
        else:
            self.help_modules()

    def _do_modules_search(self, params):
        '''Searches installed modules'''
        modules = [x for x in Framework._loaded_modules]
        if params:
            self.output(f"Searching installed modules for '{params}'...")
            modules = [x for x in Framework._loaded_modules if re.search(params, x)]
        if modules:
            self._list_modules(modules)
        else:
            self.error('No modules found.')
            self._help_modules_search()

    def _do_modules_load(self, params):
        '''Searches installed modules'''
        raise NotImplementedError

    def do_show(self, params):
        '''Shows various framework items'''
        if not params:
            self.help_show()
            return
        arg, params = self._parse_params(params)
        if arg in self._get_show_names():
            getattr(self, 'show_' + arg)()
        elif arg in self.get_tables():
            self.do_db(f"query SELECT ROWID, * FROM `{arg}`")
        else:
            self.help_show()

    def do_db(self, params):
        '''Interfaces with the workspace's database'''
        if not params:
            self.help_db()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('db'):
            return getattr(self, '_do_db_'+arg)(params)
        else:
            self.help_db()

    def _do_db_notes(self, params):
        '''Adds notes to rows in the database'''
        table, params = self._parse_params(params)
        if not table:
            self._help_db_notes()
            return
        if table in self.get_tables():
            # get rowid and note from parameters
            if params:
                arg, note = self._parse_params(params)
                rowids = self._parse_rowids(arg)
            # get rowid and note from interactive input
            else:
                try:
                    # prompt user for data
                    params = input('rowid(s) (INT): ')
                    rowids = self._parse_rowids(params)
                    note = input('note (TXT): ')
                except KeyboardInterrupt:
                    print('')
                    return
                finally:
                    # ensure proper output for resource scripts
                    if Framework._script:
                        print(f"{params}")
            # delete record(s) from the database
            count = 0
            for rowid in rowids:
                count += self.query(f"UPDATE `{table}` SET notes=? WHERE ROWID IS ?", (note, rowid))
            self.output(f"{count} rows affected.")
        else:
            self.output('Invalid table name.')

    def _do_db_insert(self, params):
        '''Inserts a row into the database'''
        table, params = self._parse_params(params)
        if not table:
            self._help_db_insert()
            return
        if table in self.get_tables():
            # validate insert_* method for table
            if not hasattr(self, 'insert_' + table):
                self.error('Cannot add records to dynamically created tables.')
                return
            columns = [x for x in self.get_columns(table) if x[0] != 'module']
            # sanitize column names to avoid conflicts with builtins in insert_* method
            sanitize_column = lambda x: '_'+x if x in ['hash', 'type'] else x
            record = {}
            # build record from parameters
            if params:
                # parse params into values by delim
                values = params.split('~')
                # validate parsed value input
                if len(columns) == len(values):
                    # assign each value to a column
                    for i in range(0,len(columns)):
                        record[sanitize_column(columns[i][0])] = values[i]
                else:
                    self.error('Columns and values length mismatch.')
                    return
            # build record from interactive input
            else:
                for column in columns:
                    try:
                        # prompt user for data
                        value = input(f"{column[0]} ({column[1]}): ")
                        record[sanitize_column(column[0])] = value
                    except KeyboardInterrupt:
                        print('')
                        return
                    finally:
                        # ensure proper output for resource scripts
                        if Framework._script:
                            print(f"{value}")
            # add record to the database
            func = getattr(self, 'insert_' + table)
            count = func(mute=True, **record)
            self.output(f"{count} rows affected.")
        else:
            self.output('Invalid table name.')

    def _do_db_delete(self, params):
        '''Deletes a row from the database'''
        table, params = self._parse_params(params)
        if not table:
            self._help_db_delete()
            return
        if table in self.get_tables():
            # get rowid from parameters
            if params:
                rowids = self._parse_rowids(params)
            # get rowid from interactive input
            else:
                try:
                    # prompt user for data
                    params = input('rowid(s) (INT): ')
                    rowids = self._parse_rowids(params)
                except KeyboardInterrupt:
                    print('')
                    return
                finally:
                    # ensure proper output for resource scripts
                    if Framework._script:
                        print(f"{params}")
            # delete record(s) from the database
            count = 0
            for rowid in rowids:
                count += self.query(f"DELETE FROM `{table}` WHERE ROWID IS ?", (rowid,))
            self.output(f"{count} rows affected.")
        else:
            self.output('Invalid table name.')

    def _do_db_query(self, params):
        '''Queries the database with custom SQL'''
        if not params:
            self._help_db_query()
            return
        try:
            results = self.query(params, include_header=True)
        except sqlite3.OperationalError as e:
            self.error(f"Invalid query. {type(e).__name__} {e}")
            return
        if type(results) == list:
            header = results.pop(0)
            if not results:
                self.output('No data returned.')
            else:
                self.table(results, header=header)
                self.output(f"{len(results)} rows returned")
        else:
            self.output(f"{results} rows affected.")

    def _do_db_schema(self, params):
        '''Displays the database schema'''
        tables = self.get_tables()
        for table in tables:
            columns = self.get_columns(table)
            self.table(columns, title=table)

    def do_script(self, params):
        '''Records and executes command scripts'''
        if not params:
            self.help_script()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('script'):
            return getattr(self, '_do_script_'+arg)(params)
        else:
            self.help_script()

    def _do_script_record(self, params):
        '''Records commands in a script file'''
        if not Framework._record:
            filename, params = self._parse_params(params)
            if not filename:
                self._help_script_record()
                return
            if not self._is_writeable(filename):
                self.output(f"Cannot record commands to '{filename}'.")
            else:
                Framework._record = filename
                self.output(f"Recording commands to '{Framework._record}'.")
        else:
            self.output('Recording is already started.')

    def _do_script_stop(self, params):
        '''Stops command recording'''
        if Framework._record:
            self.output(f"Recording stopped. Commands saved to '{Framework._record}'.")
            Framework._record = None
        else:
            self.output('Recording is already stopped.')

    def _do_script_status(self, params):
        '''Provides the status of command recording'''
        status = 'started' if Framework._record else 'stopped'
        self.output(f"Command recording is {status}.")

    def _do_script_execute(self, params):
        '''Executes commands from a script file'''
        if not params:
            self._help_script_execute()
            return
        if os.path.exists(params):
            # works even when called before Recon.start due
            # to stdin waiting for the iteractive prompt
            sys.stdin = open(params)
            Framework._script = 1
        else:
            self.error(f"Script file '{params}' not found.")

    def do_spool(self, params):
        '''Spools output to a file'''
        if not params:
            self.help_spool()
            return
        arg, params = self._parse_params(params)
        if arg in self._parse_subcommands('spool'):
            return getattr(self, '_do_spool_'+arg)(params)
        else:
            self.help_spool()

    def _do_spool_start(self, params):
        '''Starts output spooling'''
        if not Framework._spool:
            filename, params = self._parse_params(params)
            if not filename:
                self._help_spool_start()
                return
            if not self._is_writeable(filename):
                self.output(f"Cannot spool output to '{filename}'.")
            else:
                Framework._spool = codecs.open(filename, 'ab', encoding='utf-8')
                self.output(f"Spooling output to '{Framework._spool.name}'.")
        else:
            self.output('Spooling is already started.')

    def _do_spool_stop(self, params):
        '''Stops output spooling'''
        if Framework._spool:
            self.output(f"Spooling stopped. Output saved to '{Framework._spool.name}'.")
            Framework._spool = None
        else:
            self.output('Spooling is already stopped.')

    def _do_spool_status(self, params):
        '''Provides the status of output spooling'''
        status = 'started' if Framework._spool else 'stopped'
        self.output(f"Output spooling is {status}.")

    def do_shell(self, params):
        '''Executes shell commands'''
        if not params:
            self.help_shell()
            return
        proc = subprocess.Popen(params, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
        self.output(f"Command: {params}")
        stdout = proc.stdout.read()
        stderr = proc.stderr.read()
        if stdout:print(f"{Colors.O}{self.to_unicode(stdout)}{Colors.N}", end='')
        if stderr:print(f"{Colors.R}{self.to_unicode(stderr)}{Colors.N}", end='')

    def do_dashboard(self, params):
        '''Displays a summary of activity'''
        rows = self.query('SELECT * FROM dashboard ORDER BY 1')
        if rows:
            # display activity table
            tdata = []
            for row in rows:
                tdata.append(row)
            self.table(tdata, header=['Module', 'Runs'], title='Activity Summary')
            # display summary results table
            tables = self.get_tables()
            tdata = []
            for table in tables:
                count = self.query(f"SELECT COUNT(*) FROM `{table}`")[0][0]
                tdata.append([table.title(), count])
            self.table(tdata, header=['Category', 'Quantity'], title='Results Summary')
        else:
            self.output('This workspace has no record of activity.')

    def do_pdb(self, params):
        '''Starts a Python Debugger session (dev only)'''
        import pdb
        pdb.set_trace()

    #==================================================
    # HELP METHODS
    #==================================================

    def help_options(self):
        print(getattr(self, 'do_options').__doc__)
        print(f"{os.linesep}Usage: options <{'|'.join(self._parse_subcommands('options'))}> [...]{os.linesep}")

    def _help_options_set(self):
        print(getattr(self, '_do_options_set').__doc__)
        print(f"{os.linesep}Usage: options set <option> <value>{os.linesep}")

    def _help_options_unset(self):
        print(getattr(self, '_do_options_unset').__doc__)
        print(f"{os.linesep}Usage: options unset <option>{os.linesep}")

    def help_keys(self):
        print(getattr(self, 'do_keys').__doc__)
        print(f"{os.linesep}Usage: keys <{'|'.join(self._parse_subcommands('keys'))}> [...]{os.linesep}")

    def _help_keys_add(self):
        print(getattr(self, '_do_keys_add').__doc__)
        print(f"{os.linesep}Usage: keys add <name> <value>{os.linesep}")

    def _help_keys_remove(self):
        print(getattr(self, '_do_keys_remove').__doc__)
        print(f"{os.linesep}Usage: keys remove <name>{os.linesep}")

    def help_modules(self):
        print(getattr(self, 'do_modules').__doc__)
        print(f"{os.linesep}Usage: modules <{'|'.join(self._parse_subcommands('modules'))}> [...]{os.linesep}")

    def _help_modules_search(self):
        print(getattr(self, '_do_modules_search').__doc__)
        print(f"{os.linesep}Usage: modules search [<regex>]{os.linesep}")

    def _help_modules_load(self):
        print(getattr(self, '_do_modules_load').__doc__)
        print(f"{os.linesep}Usage: modules load <path>{os.linesep}")

    def help_show(self):
        options = sorted(self._get_show_names() + self.get_tables())
        print(getattr(self, 'do_show').__doc__)
        print(f"{os.linesep}Usage: show <{'|'.join(options)}>{os.linesep}")

    def help_db(self):
        print(getattr(self, 'do_db').__doc__)
        print(f"{os.linesep}Usage: db <{'|'.join(self._parse_subcommands('db'))}> [...]{os.linesep}")

    def _help_db_notes(self):
        print(getattr(self, '_do_db_notes').__doc__)
        print(f"{os.linesep}Usage: db note <table> [<rowid(s)> <note>]{os.linesep}")
        print(f"rowid(s) => ',' delimited values or '-' delimited ranges representing rowids{os.linesep}")

    def _help_db_insert(self):
        print(getattr(self, '_do_db_insert').__doc__)
        print(f"{os.linesep}Usage: db insert <table> [<values>]{os.linesep}")
        print(f"values => '~' delimited string representing column values (exclude rowid, module){os.linesep}")

    def _help_db_delete(self):
        print(getattr(self, '_do_db_delete').__doc__)
        print(f"{os.linesep}Usage: db delete <table> [<rowid(s)>]{os.linesep}")
        print(f"rowid(s) => ',' delimited values or '-' delimited ranges representing rowids{os.linesep}")

    def _help_db_query(self):
        print(getattr(self, '_do_db_query').__doc__)
        print(f"{os.linesep}Usage: db query <sql>{os.linesep}")

    def help_script(self):
        print(getattr(self, 'do_script').__doc__)
        print(f"{os.linesep}Usage: script <{'|'.join(self._parse_subcommands('script'))}> [...]{os.linesep}")

    def _help_script_record(self):
        print(getattr(self, '_do_script_record').__doc__)
        print(f"{os.linesep}Usage: script record <filename>{os.linesep}")

    def _help_script_execute(self):
        print(getattr(self, '_do_script_execute').__doc__)
        print(f"{os.linesep}Usage: script execute <filename>{os.linesep}")

    def help_spool(self):
        print(getattr(self, 'do_spool').__doc__)
        print(f"{os.linesep}Usage: spool <{'|'.join(self._parse_subcommands('spool'))}> [...]{os.linesep}")

    def _help_spool_start(self):
        print(getattr(self, '_do_spool_start').__doc__)
        print(f"{os.linesep}Usage: spool start <filename>{os.linesep}")

    def help_shell(self):
        print(getattr(self, 'do_shell').__doc__)
        print(f"{os.linesep}Usage: [shell|!] <command>{os.linesep}")

    #==================================================
    # COMPLETE METHODS
    #==================================================

    def complete_options(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('options')
        if arg in subs:
            return getattr(self, '_complete_options_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_options_list(self, text, *ignored):
        return []

    def _complete_options_set(self, text, *ignored):
        return [x for x in self.options if x.startswith(text.upper())]
    _complete_options_unset = _complete_options_set

    def complete_keys(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('keys')
        if arg in subs:
            return getattr(self, '_complete_keys_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_keys_list(self, text, *ignored):
        return []

    def _complete_keys_add(self, text, *ignored):
        return [x for x in self._get_key_names() if x.startswith(text)]
    _complete_keys_remove = _complete_keys_add

    def complete_modules(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('modules')
        if arg in subs:
            return getattr(self, '_complete_modules_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_modules_search(self, text, *ignored):
        return []

    def _complete_modules_load(self, text, *ignored):
        return [x for x in Framework._loaded_modules if x.startswith(text)]

    def complete_show(self, text, line, *ignored):
        options = sorted(self._get_show_names() + self.get_tables())
        return [x for x in options if x.startswith(text)]

    def complete_db(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('db')
        if arg in subs:
            return getattr(self, '_complete_db_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_db_insert(self, text, *ignored):
        return [x for x in sorted(self.get_tables()) if x.startswith(text)]
    _complete_db_notes = _complete_db_delete = _complete_db_insert

    def _complete_db_query(self, text, *ignored):
        return []
    _complete_db_schema = _complete_db_query

    def complete_script(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('script')
        if arg in subs:
            return getattr(self, '_complete_script_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_script_record(self, text, *ignored):
        return []
    _complete_script_execute = _complete_script_status = _complete_script_stop = _complete_script_record

    def complete_spool(self, text, line, *ignored):
        arg, params = self._parse_params(line.split(' ', 1)[1])
        subs = self._parse_subcommands('spool')
        if arg in subs:
            return getattr(self, '_complete_spool_'+arg)(text, params)
        return [sub for sub in subs if sub.startswith(text)]

    def _complete_spool_start(self, text, *ignored):
        return []
    _complete_spool_status = _complete_spool_stop = _complete_spool_start



from dicttoxml import dicttoxml
from flask import current_app, Response, jsonify, send_file, stream_with_context
from io import StringIO
from io import BytesIO
from recon.core.web.utils import add_worksheet, is_url
import os
import requests
import unicodecsv as csv
import xlsxwriter

def _jsonify(rows):
    return jsonify(rows=[dict(r) for r in rows])

def csvify(rows):
    '''Expects a list of dictionaries and returns a CSV response.'''
    if not rows:
        csv_str = ''
    else:
        s = BytesIO()
        keys = rows[0].keys()
        dw = csv.DictWriter(s, keys)
        dw.writeheader()
        dw.writerows([dict(r) for r in rows])
        csv_str = s.getvalue()
    return Response(csv_str, mimetype='text/csv')

def xmlify(rows):
    '''Expects a list of dictionaries and returns a XML response.'''
    xml = dicttoxml([dict(r) for r in rows])
    return Response(xml, mimetype='text/xml')

def listify(rows):
    '''Expects a list of dictionaries and returns a continous list of
    values from all of the provided columns.'''
    columns = {}
    for row in rows:
        for column in row.keys():
            if column not in columns:
                columns[column] = []
            columns[column].append(row[column])
    s = StringIO()
    for column in columns:
        s.write('# '+column+os.linesep)
        for value in columns[column]:
            if type(value) != str:
                value = str(value)
            s.write(value+os.linesep)
    list_str = s.getvalue()
    return Response(list_str, mimetype='text/plain')

def xlsxify(rows):
    '''Expects a list of dictionaries and returns an xlsx response.'''
    sfp = BytesIO()
    with xlsxwriter.Workbook(sfp) as workbook:
        # create a single worksheet for the provided rows
        add_worksheet(workbook, 'worksheet', rows)
    sfp.seek(0)
    return send_file(
        sfp,
        mimetype='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        as_attachment=True,
        attachment_filename=f"export.xlsx"
    )

# http://flask.pocoo.org/docs/0.12/patterns/streaming/
def proxify(rows):
    @stream_with_context
    def generate():
        '''Expects a list of dictionaries containing URLs and requests them
        through a configured proxy.'''
        # don't bother setting up if there's nothing to process
        if not rows:
            yield 'Nothing to send to proxy.'
        # disable TLS validation warning
        requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)
        # set static request options
        kwargs = {
            'headers': {
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.95 Safari/537.36',
            },
            'proxies': {
                'http': 'http://127.0.0.1:8080',
                'https': 'http://127.0.0.1:8080',
            },
            'allow_redirects': False,
            'verify': False,
        }
        # process the rows
        for row in [dict(r) for r in rows]:
            for key in row:
                url = row[key]
                msg = f"URL: {url}{os.linesep}Status: "
                if is_url(url):
                    try:
                        resp = requests.request('GET', url, **kwargs)
                        msg += f"HTTP {resp.status_code}: Successfully proxied."
                    except Exception as e:
                        msg += str(e)
                else:
                    msg += 'Error: Failed URL validation.'
                msg += os.linesep*2
                yield msg
    return Response(generate(), mimetype='text/plain')



from flask import current_app, send_file, Response, render_template
from io import BytesIO
from recon.core.web import recon
from recon.core.web.utils import columnize, add_worksheet
import xlsxwriter

def xlsx():
    '''Returns an xlsx file containing the entire dataset for the current
    workspace.'''
    sfp = BytesIO()
    with xlsxwriter.Workbook(sfp) as workbook:
        # create a worksheet for each table in the current workspace
        for table in recon.get_tables():
            rows = recon.query(f"SELECT * FROM {table}", include_header=True)
            columns = rows.pop(0)
            rows = columnize(columns, rows)
            add_worksheet(workbook, table, rows)
    sfp.seek(0)
    return send_file(
        sfp,
        mimetype='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        as_attachment=True,
        attachment_filename=f"{current_app.config['WORKSPACE']}.xlsx"
    )

def pushpin():
    google_api_key = recon.get_key('google_api')
    return Response(render_template('pushpin.html', api_key=google_api_key), mimetype='text/html')



import re
import socket
import urllib.parse
import webbrowser

class ExplicitOauthMixin(object):

    def get_explicit_oauth_token(self, resource, scope, authorize_url, access_url):
        token_name = resource+'_token'
        token = self.get_key(token_name)
        if token:
            return token
        client_id = self.get_key(resource+'_api')
        client_secret = self.get_key(resource+'_secret')
        port = 31337
        redirect_uri = f"http://localhost:{port}"
        payload = {'response_type': 'code', 'client_id': client_id, 'scope': scope, 'state': self.get_random_str(40), 'redirect_uri': redirect_uri}
        authorize_url = f"{authorize_url}?{urllib.parse.urlencode(payload)}"
        w = webbrowser.get()
        w.open(authorize_url)
        # open a socket to receive the access token callback
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('127.0.0.1', port))
        sock.listen(1)
        conn, addr = sock.accept()
        data = conn.recv(1024)
        conn.sendall('HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n\r\n<html><head><title>Recon-ng</title></head><body>Response received. Return to Recon-ng.</body></html>')
        conn.close()
        # process the received data
        if 'error_description' in data:
            self.error(urllib.parse.unquote_plus(re.search(r'error_description=([^\s&]*)', data).group(1)))
            return None
        authorization_code = re.search(r'code=([^\s&]*)', data).group(1)
        payload = {'grant_type': 'authorization_code', 'code': authorization_code, 'redirect_uri': redirect_uri, 'client_id': client_id, 'client_secret': client_secret}
        resp = self.request('POST', access_url, data=payload)
        if 'error' in resp.json():
            self.error(resp.json()['error_description'])
            return None
        access_token = resp.json()['access_token']
        self.add_key(token_name, access_token)
        return access_token
